/*
 *			L U T _ C O R R E C T . C -- Lut gamma correction stuff
 *
 *  Written  David Burgess
 *           Image Recognition Systems
 *           720 Birchwood Boulevard
 *           Birchwood Science Park
 *           Warrington
 *           Cheshire
 *           WA3 7PX
 *
 *	Copyright (c) and intellectual property rights Image Recognition Systems
 *
 *  Date:    15th January 1990
 *
 *  Modifications
 *  18 Jul 1990 ih      set top and bottom of luts to extreme values after lut_normalise -->otherwise black don't contrast with white any more !!!
 *	19 Mar 1990	dcb		lin_colour_lut, test invert, and fluor flags.
 *	14 Mar 1990	dcb		assign default if lutfile is NULL, add data corruption function
 *	13 Mar 1990	dcb		readlutfile: Allocate LUT space just before it's read in
 */
#include <stdio.h>
#include <allcontr.h>
#include <kcontrol.h>
#include <display.h>
#include <wstruct.h>

extern struct allkontr *alc;

/*
 *	D E F A U L T _ L U T -- Assign a default linear LUT
 */
struct lut_assoc *
default_lut()
{
	register unsigned int i;
	register UBYTE *rl, *gl, *bl;
	struct lut_assoc *lut;

	lut = (struct lut_assoc *)Calloc(1, sizeof(struct lut_assoc));
	lut->lut = (struct look_up_table *)Calloc(1, sizeof(struct look_up_table));

	lut->type = BRIGHT_LUT;	/* Default is brightfield */
	lut->status |= EXACT;	/* Every value is defined */
	lut->status |= 0x10;	/* .. with calculation step size of 1 */
	lut->gamma = 1.0;		/* 1.0 = linear LUT */
	rl = lut->lut->red_lut;
	gl = lut->lut->green_lut;
	bl = lut->lut->blue_lut;
	for (i=0; i<256; i++) *rl++ = *gl++ = *bl++ = 255-i;

	lut->next = NULL;
	return(lut);
}

/*
 *	F R E E L U T S -- Free of linked list of LUT structures
 */
struct lut_assoc *
Freeluts(luts)
struct lut_assoc *luts;
{
	struct lut_assoc *flut;

	while (luts) {
		flut = luts;
		luts = flut->next;
		Free(flut->lut);	/* Free the LUT */
		Free(flut);			/* Free associated info */
	}
	return(NULL);
}

/*
 *	R E A D L U T F I L E -- Read the look up tables from file
 *	File is specified in alc->lutfile.
 *	A default lut is retuned if file is NULL, or file cannot be opened
 */
struct lut_assoc *
readlutfile()
{
	FILE *lf;
	int magic, vsn, nluts, i, read_err=0;
	struct lut_assoc *luts, *lut;

	if (alc->lutfile[0] == '\0')
		luts = default_lut();
	else if ((lf = fopen(alc->lutfile, "r")) == NULL) {
		fprintf(stderr, "Can't open look-up table file %s\n", alc->lutfile);
		luts = default_lut();
	} else {
		if (((magic = getw(lf)) < 0) || (magic != LUT_MAGIC) ||
			((vsn = getw(lf)) < 0) || ((nluts = getw(lf)) < 0)) {
				fprintf(stderr, "Invalid look-up table file %s\n", alc->lutfile);
				luts = default_lut();
		} else {
			for (i=0; ((i < nluts) && (read_err == 0)); i++) {
				switch (vsn) {
				case 1:	if (i == 0) {
							lut = luts = (struct lut_assoc *)Calloc(1, sizeof(struct lut_assoc));
						} else {
							lut->next = (struct lut_assoc *)Calloc(1, sizeof(struct lut_assoc));
							lut = lut->next;
						}
 						fread(lut, sizeof(struct lut_assoc), 1, lf);
						lut->lut = (struct look_up_table *)Calloc(1, sizeof(struct look_up_table));
 						fread(lut->lut, sizeof(struct look_up_table), 1, lf);
						break;
				default: read_err = -1;
						break;
				}
			} /* end for i ... nluts */
			if (read_err < 0) { 
				fprintf(stderr, "Illegal LUT file %s\n", alc->lutfile);
				luts = Freeluts(luts);
				luts = default_lut();
			}
		}
		fclose(lf);
	}
	return(luts);
}

/*
 *	L I N _ C O L O U R _ L U T -- Apply linear LUT, and then colour overlays
 *	It is assumed that only OVERLAY1/OVERLAY2 is used for colours
 *	The background colour is picked out of 'alc'
 *	The following colours are set up:
 *	OVERLAY1 - Red, OVERLAY2 - Blue, OVERLAY1 | OVERLAY2 - Green
 */
lin_colour_lut()
{
	register int i;
	struct rgb *bg;

	bg = &(alc->bg_colour);
	if (alc->ddgschannel->invert != alc->fluor) {
		nlut();
		for (i=0; i<256; i++) {
			if ((i & (OVERLAY1|OVERLAY2)) == (OVERLAY1|OVERLAY2))
				setrgb(i, 255, 0, 0);		/* Both  = Red */
			else if (i & OVERLAY1)
				setrgb(i, 0, 255, 0);		/* OVERLAY1 = Green */
			else if (i & OVERLAY2)
				setrgb(i, 0, 255, 255);		/* OVERLAY2 = Light Blue */
		}
		setrgb(0, bg->red, bg->green, bg->blue);
	} else {
		llut();
		for (i=0; i<256; i++) {
			if ((i & (OVERLAY1|OVERLAY2)) == (OVERLAY1|OVERLAY2))
/* was 127  Joanne */		setrgb(i, 0, 200, 0);		/* Green */
			else if (i & OVERLAY1)
				setrgb(i, 255, 0, 0);		/* Red */
			else if (i & OVERLAY2)
				setrgb(i, 0, 0, 255);		/* Blue */
		}
		setrgb(0, 255 - bg->red, 255 - bg->green, 255 - bg->blue);
	}
}

/*
 *	L U T _ T Y P E --	Change the current LUT to be this type.
 *	If the type is not found, then things are left unchanged.
 *	THIS FUNCTION DOES NOT APPLY THE LUT.
 */
lut_type(type)
int type;
{
	struct lut_assoc *lut;

	for (lut = alc->luts; lut; lut = lut->next) {
		if (lut->type == type) {
			alc->lut = lut;
		}
	}
}

/*
 *	C A M E R A T Y P E _ L U T -- Apply appropriate LUT for camera in use
 */
cameratype_lut()
{
	if (alc->fluor) lut_type(FLUOR_LUT);
	else lut_type(BRIGHT_LUT);
}

/*
 *	L U T _ N O R M A L I S E --	Apply normalisation, and LUT
 *									corruption (sic) to image data
 *	Inputs: 	Object, look_up_table data
 *	Output:		Combined gamma correction/normalisation LUT.
 *				Inverted for brightfield or Flipped for fluorescent
 *	Function:	Applies combined LUT to object GREY data
 */
struct look_up_table *
lut_normalise(obj, lut)
struct object *obj;
struct look_up_table *lut;
{
	static struct look_up_table mlut;	/* Combined LUT */
	register int i;
	GREY min,max,range;
	register GREY *g;
	register UBYTE *gl, *mono_lut;
	UBYTE *rl, *bl;
	float scale;
	struct iwspace iwsp;
	struct gwspace gwsp;

	greyrange(obj, &min, &max);
	range = max - min + 1;
	gl = lut->green_lut;	/* Use just green values */
	scale = 256.0/(float)range;

	/*
	 *	Clear out the green LUT
	 */
	mono_lut = mlut.green_lut;
	for (i=0; i<256; i++) *mono_lut++ = 0;
	/*
	 *	Calculate a combined transform LUT
	 */
	mono_lut = mlut.green_lut + min;
	for (i=min; i<=max; i++)
		*mono_lut++ = 255 - *(gl + (int)(scale * (i-min)));
	/*
	 *	Now apply to the grey data
	 */
	mono_lut = mlut.green_lut;
	initgreyscan(obj,&iwsp,&gwsp);
	while (nextgreyinterval(&iwsp) == 0) {
		g = gwsp.grintptr;
		for (i=0; i<iwsp.colrmn; i++) {
			*g = *(mono_lut + *g);
			g++;
		}
	}
	/* set bottom and top to extremes ih*/
	mono_lut = mlut.green_lut;
	for (i=0; i<min; i++) *mono_lut++ = 0;
	mono_lut+=max-min+1;
	for (i=max+1; i<256; i++) *mono_lut++ = 0xff;
	/*
	 *	Create output LUT
	 */
	rl = mlut.red_lut;
	gl = mlut.green_lut;
	bl = mlut.blue_lut;
	if (alc->fluor) {
		gl += 255;
		for (i=0; i<=255; i++) *rl++ = *bl++ = *gl--;
		gl = mlut.green_lut;
		bl = mlut.blue_lut;
		for (i=0; i<=255; i++) *gl++ = *bl++;
	} else {
		for (i=0; i<=255; i++) *rl++ = *bl++ = *gl++ = 255 - *gl;
	}
	return(&mlut);
}

