/*
 * flex.h			Jim Piper			23-10-90
 *
 * Structures and bit-definitions (and some documentation) for
 * the flexible screen system.
 *
 * MODIFICATIONS
 * -------------
 * 31-01-91 jimp	Added DigType to flex_status structure.
 * 30-01-91 jimp	Added "modified" to flex_status structure, to
 *			indicate necessity for writing to disc.
 * 29-01-91 jimp	Added struct flex_poly_plist for polygon (arrow
 *			or box) width and type.
 * 19-01-91 jimp	Added "alltoggle" to status (switches enhancements
 *			from selected group to all objects, or back).
 * 20-12-90 BH		Added "Phase", "SelectObj", "Frames", "TrackFuncs",
 *			"ObjPickedUp", "KaryId" and "SelectWayup" to 
 *			flex_status structure
 * 18-12-90 jimp	Added "boxes" to status, "nameids" to id.
 * 17-12-90 jimp	Tightened up the identification things with
 *			struct flex_header *header
 * 06-12-90 jimp	Added struct flex_status to flexible_screen.
 * 26-11-90 jimp	Added "current_obj" to struct flexible_screen.
 *
 *------------------------------------------------------------------------------
 *				DOCUMENTATION
 *				-------------
 * The flexible screen consists of an identifier and a compound Woolz object.
 *
 * The compound object is the root of a hierarchy or tree that comprises
 * an arbitrary mix of the following types, all based on Woolz "compounda"
 * (array compound) object:
 *	(1) leaves - essentially individual objects (chromosomes, pointers,
 *		textual comments).  However, a leaf is itself a 2-element
 *		compound object, comprising the individual object and its
 *		own private "pframe" (in the first slot) for display purposes.
 *	(2) node groups - comprising a compound object whose elements may
 *		either be leaves or other non-leaf groups.  A non-leaf
 *		group ALWAYS contains a "struct irect" that specifies in
 *		DIGS coordinates the bounding right-rectangle of all the
 *		group members.  A leaf is distinguished from a node by the
 *		first element - always a "pframe" in a leaf, always an "irect"
 *		in a node.
 *	(3) NULL objects - where groups or leaves have for one reason or
 *		another been moved or removed.
 * In each of cases (1) and (2), the compounda "assoc" member is used
 * to point to the containing node (taken to include the root).
 *
 * Image objects -- i.e. chromosomes -- must be COPIED since the originals
 * may be freed when processing a cell is finished, but the flexible screen
 * must keep the objects through the processing of further cells.
 *
 * Because IRS do not yet have linkcounts in (top-level) objects, some
 * care is required with storage management in compound objects ......
 *
 * The bounding right-rectangle vertex order is (low y, low x),
 * (high y, low x), (high y, high x), (low y, high x).
 *
 * Routines are often recursive.  Most are named flex_xxx_sc if they
 * operate on the entire flexible screen, and flex_xxx_nd if they act
 * recursively on the nodes of the screen.
 *------------------------------------------------------------------------------
 */

struct flexible_screen {
	struct flex_id *id;	/* identifier of this screen */
	struct compounda *cao;	/* list of groups */
	struct flex_status *stat;
};

struct flex_status {
	struct compounda *current_obj;	/* object presently identified */
	struct chromosome *SelectObj; 	/* obj selected in source screen */
	struct pframe Frames[4];	/* frames for the 4 quadrants */
	struct kident KaryId;		/* karyotype id for selection */
	int (*TrackFuncs[4])();		/* tracking functions in 4 quadrants */
	int Phase;			/* present phase, 0 segment else kary */
	int ObjPickedUp;		/* has an object been picked up? */
	int SelectWayUp;		/* 1 selected obj normal, -1 inverted */
	int DigType;			/* DIGTV or DIGFIP */
	int scale;			/* display scale of image objects */
	int dx, dy;			/* display origin of flexible screen */
	int boxes;			/* toggle for boxes on or off */
	int alltoggle;			/* toggle for groups/all enhancements */
	int modified;
};

struct flex_id {
	struct flex_header *header;	/* note this IS a pointer */
	struct cyto_file nameids;	/* note this NOT a pointer, cf. "allkontr" */
	char filename[32];
};

struct flex_poly_plist {
	SMALL size;		/* of this structure - ESSENTIAL for I/O */
	char linewidth;
	char polytype;		/* 1 an arrow, 2 a box */
};

/*
 * The two screen styles (top-left quadrant, whole screen)
 * are described by the following parameters (see flex_frames_small/big)
 */
#define X_SMALL	0
#define X_LARGE	0
#define Y_SMALL	2300	/* Will be wrong for a non-Virtuoso */
#define Y_LARGE	0
#define S_SMALL	4
#define S_LARGE	8


