/*
 * flexquads.c		Jim Piper and Bill @ MRC HGU		04-12-90
 *
 * This file contains functions for initialising the flexible screen quadrants
 * and transforming coordinates between the quadrants.
 *
 * MODIFICATIONS
 * -------------
 * 20-12-90 BH          Flex globals now in flex_status structure and parameters
 *			of functions flex_Shift(From)|(To)Quad(X)|(Y) have 
 *			been changed to take a pointer to appropriate quad
 *			frame instead of the quadrant number.
 *
 */

#include <stdio.h>
#include <kyy.h>
#include <chromanal.h>
#include <kcontrol.h>
#include <flex.h>

int
		flex_ScaleFromQuad(),
		flex_ScaleToQuad(),
		flex_ShiftFromQuadX(),
		flex_ShiftFromQuadY(),
		flex_ShiftToQuadX(),
		flex_ShiftToQuadY(),
		flex_WhichQuadrant();
void		flex_InitQuadrants(),
		flex_SplitScreen();


extern struct kcontrol *kcont;
static int	flex_SplitX,
		flex_SplitY;

/************************************************************************
* Function:	flex_InitQuadrants					*
* Returns:	void							*
* Purpose:	Initialise the quadrants by dividing the display up	*
*		into four quadrants and creating a pframe for each one.	*
*		Quadrants are:   	-------------------		*
*					| Quad 0 | Quad 1 |		*
*					| (Flex) | (Src0) |		*
*					---------+---------		*
*					| Quad 2 | Quad 3 |		*
*					| (Src2) | (Src1) |		*
*					-------------------		*
*		Also make the kcontrol pframe a bit more sensible.	*
* Parameters:	-							*
************************************************************************/
void		flex_InitQuadrants()
{
  register int Quad;
  register struct pframe *Frame;

  flex_SplitX = (alc->ddgschannel->lineln << alc->ddgschannel->xshift) >> 1;
  flex_SplitY = (alc->ddgschannel->lineno << alc->ddgschannel->yshift) >> 1;
  Frame = alc->flex_screen->stat->Frames;
  for(Quad = 0; Quad < 4; ++Quad)
  {
    Frame->type = 60;
    Frame->scale = 4;
    switch(Quad)
    {
      case 0:
	Frame->dx = 0;
	Frame->dy = flex_SplitY;
	break;
      case 1:
	Frame->dx = flex_SplitX;
	Frame->dy = flex_SplitY;
	break;
      case 2:
	Frame->dx = 0;
	Frame->dy = 0;
	break;
      case 3:
	Frame->dx = flex_SplitX;
	Frame->dy = 0;
	break;
    }
    Frame->ox = 0;
    Frame->oy = 0;
    Frame->ix = 1;
    Frame->iy = 1;
    Frame->func = (int(*)() )0;
    ++Frame;
  }
  Frame = kcont->f;
  Frame->type = 60;
  Frame->scale = 8;
  Frame->dx = 0;
  Frame->dy = 0;
  Frame->ox = 0;
  Frame->oy = 0;
  Frame->ix = 1;
  Frame->iy = 1;
  Frame->func = (int(*)() )0;
}

/************************************************************************
* Function:	flex_WhichQuadrant					*
* Returns:	int:			quadrant number [0-3]		*
* Purpose:	Find which quadrant the given coordinate pair lies in.	*
* Parameters:	int xCoOrd:		given ddgs x coordinate		*
*		int yCoOrd:		given ddgs y coordinate		*
************************************************************************/
int		flex_WhichQuadrant(xCoOrd, yCoOrd)
int		xCoOrd,
		yCoOrd;
{
  int		Quad;

  if(xCoOrd > flex_SplitX)
    Quad = 1;
  else
    Quad = 0;
  if(yCoOrd < flex_SplitY)
    Quad += 2;
  return(Quad);
}

/************************************************************************
* Function:	flex_SplitScreen					*
* Returns:	void							*
* Purpose:	Split the screen into quadrants with a cross.		*
* Parameters:	-							*
************************************************************************/
void		flex_SplitScreen()
{
  lwidth(3);
  colour(255 & ~(OVERLAY1 | OVERLAY2));
  moveto(0, flex_SplitY);
  lineto(flex_SplitX + flex_SplitX, flex_SplitY);
  moveto(flex_SplitX, 0);
  lineto(flex_SplitX, flex_SplitY + flex_SplitY);
  lin_colour_lut();
  lwidth(1);
}

/************************************************************************
* Function:	flex_ScaleToQuad					*
* Returns:	int:			scaled parameter		*
* Purpose:	Scale given parameter to fit a quadrant of display.	*
* Parameters:	int Given:		given parameter			*
************************************************************************/
int		flex_ScaleToQuad(Given)
register int	Given;
{
  return(Given>>1);
}

/************************************************************************
* Function:	flex_ScaleFromQuad					*
* Returns:	int:			scaled parameter		*
* Purpose:	Scale given parameter from quadrant to whole display.	*
* Parameters:	int Given:		given parameter			*
************************************************************************/
int		flex_ScaleFromQuad(Given)
register int	Given;
{
  return(Given<<1);
}

/************************************************************************
* Function:	flex_ShiftToQuadX					*
* Returns:	int:			scaled parameter		*
* Purpose:	Scale given x parameter and shift to display quadrant.	*
* Parameters:	int Given:		given parameter			*
*		pframe *Frame:		required quadrant frame		*
************************************************************************/
int		flex_ShiftToQuadX(Given, Frame)
register int	Given;
register struct pframe *Frame;
{
  return((Given >> 1) + Frame->dx);
}

/************************************************************************
* Function:	flex_ShiftToQuadY					*
* Returns:	int:			scaled parameter		*
* Purpose:	Scale given y parameter and shift to display quadrant.	*
* Parameters:	int Given:		given parameter			*
*		pframe *Frame:		required quadrant frame		*
************************************************************************/
int		flex_ShiftToQuadY(Given, Frame)
register int	Given;
register struct pframe *Frame;
{
  return((Given >> 1) + Frame->dy);
}

/************************************************************************
* Function:	flex_ShiftFromQuadX					*
* Returns:	int:			scaled parameter		*
* Purpose:	Shift given x parameter and scale to whole display.	*
* Parameters:	int Given:		given parameter			*
*		pframe *Frame:		source quadrant frame		*
************************************************************************/
int		flex_ShiftFromQuadX(Given, Frame)
register int	Given;
register struct pframe *Frame;
{
  return((Given - Frame->dx) << 1);
}

/************************************************************************
* Function:	flex_ShiftFromQuadY					*
* Returns:	int:			scaled parameter		*
* Purpose:	Shift given y parameter and scale to whole display.	*
* Parameters:	int Given:		given parameter			*
*		pframe *Frame:		source quadrant frame		*
************************************************************************/
int		flex_ShiftFromQuadY(Given, Frame)
register int	Given;
register struct pframe *Frame;
{
  return((Given - Frame->dy) << 1);
}
