/*
 *		Z O O M B O X . C  --
 *
 *
 *  Written: Clive A Stubbings
 *           Image Recognition Systems
 *           720 Birchwood Boulevard
 *           Birchwood Science Park
 *           Warrington
 *           Cheshire
 *           WA3 7PX
 *
 *	Copyright (c) and intellectual property rights Image Recognition Systems
 *
 *  Date:    16th October 1987
 *
 *	Based on an idea by Denis Rutovitz	Nov 1986, Modified by Jim Piper
 *						Medical Research Council Cytogenetics Unit
 *						Edinburgh
 *  Modifications
 *
 *	04 Apr 1990		dcb			Different window shade for fluorescent
 *	21 Feb 1989		SEAJ		Changed setbframe and dispbig to add an extra 
 *								parameter ignore to allow plist->dispor to be
 *								ignored.
 *  31 Jan 1989		SEAJ		Added extra debug statements.
 *	26 Oct 1987		CAS			Tweaking scaling
 */
#include <stdio.h>
#include <allcontr.h>
#include <wstruct.h>
#include <chromanal.h>
#include <display.h>

#define MIN(a,b) (((a) > (b)) ? (b) : (a))
#define ABS(a) (((a) > 0) ? (a) : -(a))

#define BRIGHTWINDSHADE	32	/* Window background colour - brightfield*/
#define FLUORWINDSHADE	92	/* Window background colour - fluorescence*/
#define MARGIN		160	/* Margin around object (including border) */
#define EDGEGUARD	200	/* to keep things away from limits of access */

extern struct allkontr *alc;
extern FILE *debug_channel;

static char *savewindow;		/* Pointer to saved window */


/*
 *	D I S P B I G  --	Display an object in a zoomed window.
 *
 *		phase		0	karyotype display
 *		howbig			how much bigger than it is now to display it
 *
 */
struct pframe *dispbig(obj,oldframe,phase,howbig, ignore)
struct object *obj;
struct pframe *oldframe;
int phase,howbig, ignore;
{
	static	struct pframe bigframe;
	int		stat, scalefactor;
	int		curleft, curright, curbot, curtop;
	int		winleft, winright, winbot, wintop;
/*
 *	Calculate size of current box and ensure right > left etc
 */
	curleft = frdigsxpos(obj->idom->kol1,oldframe);
	curright = frdigsxpos(obj->idom->lastkl,oldframe);
	if (curleft > curright) swop (&curleft,&curright);
	curbot = frdigsypos(obj->idom->line1,oldframe);
	curtop = frdigsypos(obj->idom->lastln,oldframe);
	if (curbot > curtop) swop (&curbot,&curtop);
/*
 *	Now calculate the size we want the new box to be allowing a sensible
 *	margin arround the object (+ space to draw the border)
 *
 *	Note - cur values are multiplied by oldframe scale. If x + y increments
 *			are same we can get more variation in scale factors by keeping
 *			the common factor out of the overall scalefactor calc.
 */
	scalefactor = 1 + MIN(ABS(howbig/oldframe->ix),ABS(howbig/oldframe->iy));
	if (alc->debug)
		fprintf(debug_channel,"Zoombox, oldframe ox %d oy %d ix %d, iy %d, scale %d dx %d dy %d\n",
		oldframe->ox, oldframe->oy, oldframe->ix, oldframe->iy, oldframe->scale, oldframe->dx, oldframe->dy);
	Alloc_wait(0);			/* Disable memory wait in case request fails */
	do {
		scalefactor--;
		stat = 0;
		winleft = winbot = 0;
		winright = ((scalefactor * (curright - curleft)) / oldframe->scale) + MARGIN;
		wintop = ((scalefactor * (curtop - curbot)) / oldframe->scale) + MARGIN;
		if (!placewindow(alc->ddgschannel,&winleft,&winright,&winbot,&wintop,curleft,curright,curbot,curtop))
			continue;
		stat = clearbox(winleft,winbot,winright-winleft,wintop-winbot);
	} while (!stat && (scalefactor > 1));
	Alloc_wait(1);
	if (!stat) {
		clearbox(winleft,winbot,winright-winleft,wintop-winbot);
	}
/*
 *	Sort out the new object frame
 */
	setbframe(obj, phase, &bigframe, ignore);
	bigframe.scale = scalefactor;
	bigframe.ix = oldframe->ix;
	bigframe.iy = oldframe->iy;
	bigframe.dx = winleft + MARGIN/2;
	bigframe.dy = winbot + MARGIN/2;
	if (alc->debug)
		fprintf(debug_channel,"Zoombox, bigframe ox %d oy %d ix %d, iy %d, scale %d dx %d dy %d\n",
		bigframe.ox, bigframe.oy, bigframe.ix, bigframe.iy, bigframe.scale, bigframe.dx, bigframe.dy);
/*
 *	Finally draw the border round the window + put the enlarged object in it
 */
	colour(OVERLAY2);
	lwidth(3);
	moveto(winleft+16 ,winbot+16);
	lineto(winright-32,winbot+16);
	lineto(winright-32,wintop-32);
	lineto(winleft+16 ,wintop-32);
	lineto(winleft+16 ,winbot+16);
	lwidth(1);
	colour(GREYCOLS);
	picframe(obj,&bigframe);
	return(&bigframe);
}

/*
 *	S E T B F R A M E  --	Set the object-origin appropriately if phase=0 (karyotype display)
 *
 */
static setbframe(obj,phase,f, ignore)
struct chromosome *obj;
int	phase, ignore;
struct pframe *f;
{
	if ((ignore) || ((phase == 0) && (obj->plist->dispor >= 0))) {
		f->ox = obj->idom->kol1;
		f->oy = obj->idom->line1;
	} else {
		f->ox = obj->idom->lastkl;
		f->oy = obj->idom->lastln;
	}
}


/*
 *	P L A C E W I N D O W  --	Work out where to put window on screen
 *
 *	Window is placed in middle of largest free space avoiding origonal object
 *	if possible. Otherwise object overwritten if necessary
 *
 *	Note:	winleft + winbot are initially ZERO hence winright + wintop
 *			are effectively width + hight
 */
placewindow(dg,winleft,winright,winbot,wintop,curleft,curright,curbot,curtop)
DDGS	*dg;
int		*winleft,*winright,*winbot,*wintop,curleft,curright,curbot,curtop;
{
	int		fsminy, fsmaxy, fsminx, fsmaxx;
	fsminy = dg->miny<<dg->yshift;
	fsmaxy = dg->maxy<<dg->yshift;
	fsminx = dg->minx<<dg->xshift;
	fsmaxx = dg->maxx<<dg->xshift;
/*
 *	Check there is room for window on screen
 */
	if ((*winright > (fsmaxx - fsminx)) || (*wintop > (fsmaxy - fsminy)))
		return(0);
/*
 *	Check current object is on screen - if not stick enlarged box
 *	in middle of screen
 */
	if ((curleft > fsmaxx) || (curright < fsminx) || (curbot > fsmaxy) || (curtop < fsminy)) {
		*winbot = (fsminy + fsmaxy - *wintop) / 2;
		*wintop += *winbot;
		*winleft = (fsminx + fsmaxx - *winright) / 2;
		*winright += *winleft;
		return(1);
	}
/*
 *	Set window position vertically
 */
	if ((curbot - fsminy) > (fsmaxy - curtop)) {		/* Bottom half */
		if ((curbot - fsminy) > (*wintop + 2 * EDGEGUARD)) {
			*winbot = (curbot + fsminy - *wintop) / 2;
			*wintop += *winbot;
		} else {
			*winbot = EDGEGUARD;
			*wintop += EDGEGUARD;
		}
	} else {											/* top half */
		if ((fsmaxy - curtop) > (*wintop + 2 * EDGEGUARD)) {
			*winbot = (fsmaxy + curtop - *wintop) / 2;
			*wintop += *winbot;
		} else {
			*winbot = fsmaxy - *wintop - EDGEGUARD;
			*wintop = fsmaxy - EDGEGUARD;
		}
	}
/*
 *	Set window position horizontally
 */
	if ((curleft - fsminx) > (fsmaxx - curright)) {		/* Left half */
		if ((curleft - fsminx) > (*winright + 2 * EDGEGUARD)) {
			*winleft = (curleft + fsminx - *winright) / 2;
			*winright += *winleft;
		} else {
			*winleft = EDGEGUARD;
			*winright += EDGEGUARD;
		}
	} else {											/* right half */
		if ((fsmaxx - curright) > (*winright + 2 * EDGEGUARD)) {
			*winleft = (fsmaxx + curright - *winright) / 2;
			*winright += *winleft;
		} else {
			*winleft = fsmaxx - *winright - EDGEGUARD;
			*winright = fsmaxx - EDGEGUARD;
		}
	}
	return(1);
}


/*
 *	S W O P   --   Swop two ints
 *
 */
static swop(a,b)
register int *a,*b;
{
	register c = *b;
	*b = *a;
	*a = c;
}


/*
 *	C L E A R B O X  --	Save + clear window
 *
 */
clearbox(windowxpos,windowypos,windowxdel,windowydel)
{
	int windshade;
	char *wclearsave();

	windshade = alc->fluor ? FLUORWINDSHADE : BRIGHTWINDSHADE;
	moveto(windowxpos, windowypos);
	savewindow = wclearsave(windowxdel, windowydel, windshade);
	return(savewindow);
}

/*
 *	P U T B A C K W I N D O W  --	restore things as they were
 *
 */
putbackwindow()
{
	wrestore(savewindow,0);
}
