/*
 *	A F E _ C T R L . C -- Functions for controlling new analogue front end
 *
 *
 *  Written: David Burgess
 *           Image Recognition Systems
 *           720 Birchwood Boulevard
 *           Birchwood Science Park
 *           Warrington
 *           Cheshire
 *           WA3 7PX
 *
 *	Copyright (c) and intellectual property rights Image Recognition Systems (1989)
 *
 *	Notes:	All afe_... functions report the value to which they are set
 *	when called with a -1 parameter (enquire function)
 *
 *  Date:    23rd January 1989
 *
 *  Modifications
 * 23 May 1989	dcb		Return index from get_gain NOT control bits
 * 22 May 1989	dcb 	Send ctrl_bits to the h/ware NOT dr !
 * 16 May 1989	dcb		density_range, use supplied dr as gv[] index to give
 *						increasing gain for density range 0-6
 * 13 Feb 1989	dcb		Only init afe from density_range, if present.
 * 10 Feb 1989	dcb		Remove variable upper limit to dynamic range
 *						and reduce top val %age, remove "use next lower range"
 */
#include <stdio.h>
#include <allcontr.h>
#include <fipcomm.h>
#include <afe_ctrl.h>

/*
 *	External things to import
 */
extern struct allkontr *alc;
extern FILE *debug_channel;

/*
 * Gain ranges, max. dynamic range available,
 * and control bits to give this gain.
 * Note: Gains are specifically arranged to be in order of increasing value
 * Eventually max dynamic range will be set by calculation
 * Bit 2 is set by sub_unit 2, command 2, or 6 (AUTOFOCUS),
 * Bits 1, 0 are set by sub_unit 0, command 3
 */
static struct gain_range gv[] = {
/*						   2 1 0 */
	{ 25, 7 },	/* 1 1 1 - old upper limit 228 */
 	{ 35, 5 },	/* 1 0 1 - old upper limit 220 */
	{ 46, 6 },	/* 1 1 0 - old upper limit 212 */
	{ 55, 3 },	/* 0 1 1 - old upper limit 206 */
	{ 77, 1 },	/* 0 0 1 - old upper limit 190 */
	{ 90, 4 },	/* 1 0 0 - old upper limit 180 */
	{104, 2 },	/* 0 1 0 - old upper limit 170 */
	{201, 0 }	/* 0 0 0 - old upper limit 99  */
	};

static struct analogue_front_end *afe=NULL;
static int afe_errno = 0;	/* Error number */
static int wtm = 1;			/* Write to miproc */

/*
 * G E T _ G A I N - Set actual gain value, and return
 * the INDEX into gv array (This is the value the user sees as density range)
 * All gain values in gv are 10 times actual value due to compiler limitation
 */
int
get_gain(top_val, calib_gain, avail_gain)
int top_val;		/* Top value in histogram */
float calib_gain;	/* Gain at which histo. scan was done */
float *avail_gain;	/* To return available gain setting to use */
{
	int index; /* for indexing into gv array */
	float reqd_gain;

	index = find_index(calib_gain); /* Find index in gv corresp. to this gain */
	if (alc->debug)
		fprintf(debug_channel, "Gain index = %d, top_val = %d\n",
			index, top_val);
	reqd_gain = ((float)MAX_RANGE/(float)top_val) * calib_gain;

	/*
	 * Now choose the available setting as the gain below this value
	 * Note: Gains in gv are arranged to be in order of increasing value
	 */
	if (alc->debug) {
		fprintf(debug_channel, "No. of ranges = %d\n",
			(sizeof(gv)/sizeof(struct gain_range)));
		fprintf(debug_channel, "calib_gain = %f\n", calib_gain);
		fprintf(debug_channel, "reqd_gain = %f\n", reqd_gain);
	}
	for (index=0; index < (sizeof(gv)/sizeof(struct gain_range)); index++)
		if (reqd_gain < ((float)gv[index].gain/10)) break; /* Found it */

	index--;	/* Select lower range */
	if (alc->debug)	fprintf(debug_channel, "index = %d\n", index);
#ifdef JNK
	/*
	 * We don't want to overflow the dynamic range available, so
	 * Now check that (top value * new gain incr.) < max dynamic range avail.
	 * new gain incr. = (this gain setting)/calib_gain
	 */
	/*find_index(*avail_gain);	*/
	while ((index>=0) &&
		((top_val*((gv[index].gain/10)/calib_gain)) > gv[index].max_range)) {
			if (alc->debug)
				fprintf(debug_channel, "top_val = %d, gain = %d, cal gain = %f,max range = %d\n",
				top_val, gv[index].gain, calib_gain, gv[index].max_range);
			fprintf(stderr, "Available gain setting too high using next lower\n");
			index--;	/* So use next lower, and try again */
	}
	if (index == -1) {
		fprintf(stderr, "get_gain: Can't set gain low enough\n");
		return(-1);	/* Error: Can't set gain low enough */
	}
#endif

	if (alc->debug)
		fprintf(debug_channel, "top_val = %d, gain = %d, cal gain = %f\n",
			top_val, gv[index].gain, calib_gain);
	*avail_gain = (float)gv[index].gain/10;	/* Return gain to be used */
	if (alc->debug)
		fprintf(debug_channel, "Use gain %f\n", *avail_gain);
	return(index);	/* index to gv array for this gain */
} /* end get_gain */

/*
 * A F E _ I N I T -- Initialise analogue front end card
 */
afe_init()
{
	if (afe) return(AFE_CI); /* Already initialised */
	if (alc->debug) fprintf(debug_channel, "Initialising afe\n");
	if ((afe = (struct analogue_front_end *)Calloc(1, sizeof(struct analogue_front_end))) == NULL) {
		fprintf(stderr, "Can't init afe\n");
		return(AFE_CI);
	}
	afe_errno = 0;					/* No error yet */
	wtm = 0;						/* Don't write until all set */
	afe_test(VIDEO_IP, VIDEO);		/* Select video input on video card	*/
	afe_test(VIDEO_IP, AUTO_FOCUS);	/* Select video input on a/f card	*/
	afe_test_level(0, VIDEO);		/* Test level = 0 for video card	*/
	afe_test_level(0, AUTO_FOCUS);	/* Test level = 0 for video card	*/
	afe_int_den(DENSITY);			/* Density mode (No effect at present) */
	afe_interp(INTERP_OFF, VIDEO);	/* Interpolator off video card		*/
	afe_interp(INTERP_OFF, AUTO_FOCUS);	/* Interpolator off a/f card	*/
	afe_gain(DEF_GAIN, VIDEO);		/* Default gain, video card			*/
	afe_gain(DEF_GAIN, AUTO_FOCUS);	/* Default gain, a/f card			*/
	wtm = 1;
	write_afe();					/* Hit miproc once only */
	return(afe_errno);				/* Error number, or zero if no error */
} /* end afe_init */

/*
 * A F E _ G A I N -- Set gain
 * Gain = -1, report current setting, otherwise is actual gain
 */
int
afe_gain(gain, board)
float gain;
int board;
{
	int index, stat;
	if (afe_errno) return(afe_errno);
 	if (afe == NULL)
		if ((stat = afe_init()) < 0) {
			afe_errno = stat;
			return(stat);
		}
	if (gain < 0) return((int)(afe->gain*10));	/* Current gain		*/
	afe->gain = gain;		/* Keep a record */
	index = find_index(gain);
	if (alc->debug)
		fprintf(debug_channel, "afe_gain: Set gain %f, index=%d, ctrl=%d\n",
			gain, index, gv[index].control_bits);
	switch (board) {
	case VIDEO:	 /* Access video card */
		afe->su_cm_0_3 &= ~EXIST_GAIN;		/* Clear previous	*/
		afe->su_cm_0_3 |= gv[index].control_bits & EXIST_GAIN;	/* put in new */
		write_su_cm(0, 3, afe->su_cm_0_3);	/* Set in PPU		*/
		afe->su_cm_2_2 &= ~ADD_GAIN;		/* Clear previous	*/
		afe->su_cm_2_2 |= gv[index].control_bits & ADD_GAIN;	/* put in new */
		write_su_cm(2, 2, afe->su_cm_2_2);	/* Set in PPU		*/
		break;								/* All done			*/
	case AUTO_FOCUS:	/* Access Auto-focus card */
		afe->su_cm_0_3 &= ~EXIST_GAIN;		/* Clear previous	*/
		afe->su_cm_0_3 |= gv[index].control_bits & EXIST_GAIN;	/* put in new */
		write_su_cm(0, 3, afe->su_cm_0_3);	/* Set in PPU		*/
		afe->su_cm_2_6 &= ~ADD_GAIN;		/* Clear previous	*/
		afe->su_cm_2_6 |= gv[index].control_bits & ADD_GAIN;	/* put in new */
		write_su_cm(2, 6, afe->su_cm_2_6);	/* Set in PPU		*/
		break;								/* All done			*/
	default:
		return(afe_errno = AFE_PE);			/* Parameter error	*/
		break;
	} /* End switch board */
	return(0);
} /* end afe_gain*/

/*
 * A F E _ I N T _ D E N -- Set intensity/density mode
 * int_den = 0 for Density mode
 *		   = 1 for Intensity mode
 */
afe_int_den(int_den)
int int_den;
{
	int stat;
 	if (afe == NULL)
		if ((stat = afe_init()) < 0) {
			afe_errno = stat;
			return(stat);
		}
	/* Null function for present */
	return(0);
} /* end afe_int_den */

/* 
 * A F E _ I N T E R P -- Set/unset interpolate mode
 * interp = INTERP_ON  (0) for interpolator on
 *		  = INTERP_OFF (8) for interpolator off (Bit 3 set)
 *		  = -1 Give current value
 */
afe_interp(interp, board)
int interp, board;
{
	int stat;
	if (afe_errno) return(afe_errno);
 	if (afe == NULL)
		if ((stat = afe_init()) < 0) {
			afe_errno = stat;
			return(stat);
		}
	switch (board) {
	case VIDEO:	 /* Access video card */
		if (interp < 0)
			return(afe->su_cm_2_2 | INTERP);	/* Current mode		*/
		afe->su_cm_2_2 &= ~INTERP;				/* Clear previous	*/
		afe->su_cm_2_2 |= interp & INTERP;		/* put in new		*/
		write_su_cm(2, 2, afe->su_cm_2_2);		/* Set in PPU		*/
		break;									/* All done			*/
	case AUTO_FOCUS:	/* Access Auto-focus card */
		if (interp < 0)
			return(afe->su_cm_2_6 | INTERP);	/* Current mode		*/
		afe->su_cm_2_6 &= ~INTERP;				/* Clear previous	*/
		afe->su_cm_2_6 |= interp & INTERP;		/* put in new		*/
		write_su_cm(2, 6, afe->su_cm_2_6);		/* Set in PPU		*/
		break;									/* All done			*/
	default:
		return(afe_errno = AFE_PE);				/* Parameter error	*/
		break;
	} /* End switch board */
	return(0);
} /* end afe_interp */

/*
 * A F E _ T E S T -- Set test mode (video/ramp/level)
 * mode = -1 Give current mode for board
 *		=  RAMP_IP (0) Select Ramp mode test
 *		=  VIDEO_IP (1) Select Video input
 *		=  LEVEL_IP (2) Select constant level test
 */
afe_test(mode, board)
int mode, board;
{
	int stat;
 	if (afe == NULL)
		if ((stat = afe_init()) < 0) {
			afe_errno = stat;
			return(stat);
		}
	if (afe_errno) return(afe_errno);
	switch (board) {
	case VIDEO:	/* Access video card */
		if (mode < 0)
			return(afe->su_cm_2_2 | TEST_SIG);	/* Current mode		*/
		afe->su_cm_2_2 &= ~TEST_SIG;			/* Clear previous	*/
		afe->su_cm_2_2 |= mode & TEST_SIG;		/* put in new		*/
		write_su_cm(2, 2, afe->su_cm_2_2);		/* Set in PPU		*/
		break;									/* All done			*/
	case AUTO_FOCUS:	/* Access Auto-focus card */
		if (mode < 0)
			return(afe->su_cm_2_6 | TEST_SIG);	/* Current mode		*/
		afe->su_cm_2_6 &= ~TEST_SIG;			/* Clear previous	*/
		afe->su_cm_2_6 |= mode & TEST_SIG;		/* put in new		*/
		write_su_cm(2, 6, afe->su_cm_2_6);		/* Set in PPU		*/
		break;									/* All done			*/
	default:
		return(afe_errno = AFE_PE);				/* Parameter error	*/
		break;
	} /* End switch board */
	return(0);
} /* end afe_test */

/*
 * A F E _ T E S T _ L E V E L -- Set test constant level value
 * value	= -1  Give current value
 *			= +ve Value to set
 */
afe_test_level(value, bored)
int value, bored;
{
	int stat;
	if (afe_errno) return(afe_errno);
 	if (afe == NULL)
		if ((stat = afe_init()) < 0) {
			afe_errno = stat;
			return(stat);
		}
	switch (bored) {
	case VIDEO:	 /* Access video card */
		if (value < 0) return(afe->su_cm_2_3);	/* Current val	*/
		afe->su_cm_2_3 = value;					/* Set new val	*/
		write_su_cm(2, 3, afe->su_cm_2_3);		/* Set in PPU	*/
		break;									/* All done		*/
	case AUTO_FOCUS:	/* Access Auto-focus card */
		if (value < 0) return(afe->su_cm_2_7);	/* Current val	*/
		afe->su_cm_2_7 = value;					/* Set new val	*/
		write_su_cm(2, 7, afe->su_cm_2_7);		/* Set in PPU	*/
		break;									/* All done		*/
	default:
		return(afe_errno = AFE_PE);				/* Parameter error	*/
		break;
	} /* End switch board */
	return(0);
} /* end afe_test_level */

/*
 *	D E N S I T Y _ R A N G E -- Set density range (gain) according to bit
 *	pattern. density_range is used as an index to look up the actual control
 *	bits since these are weird (Thanks NEL!)
 *	Also tests if new afe present. This fn. is similar to afe_gain
 *	NOTE:
 *	Gain is set for video card only.
 *	The afe->gain field is also set
 */
density_range(dr)
int dr;
{
	int ctrl_bits, i, stat=0;
 	if ((afe == NULL) && (alc->config & CONF_FIPAL))
		if ((stat = afe_init()) < 0) {
			afe_errno = stat;
			return(stat);
		}
	ctrl_bits = gv[dr].control_bits;
	if (alc->debug)
		fprintf(debug_channel, "density_range: Set gain control bits = %d\n",
				ctrl_bits);
	if (alc->config & CONF_FIPAL) {					/* New card present */
		afe->su_cm_0_3 &= ~EXIST_GAIN;				/* Clear previous	*/
		afe->su_cm_0_3 |= (ctrl_bits & EXIST_GAIN);	/* put in new bits	*/
		stat = write_su_cm(0, 3, afe->su_cm_0_3);	/* Set in PPU		*/
		if (stat != 0) return(stat);
		afe->su_cm_2_2 &= ~ADD_GAIN;				/* Clear previous	*/
		afe->su_cm_2_2 |= ctrl_bits & ADD_GAIN;		/* put in new bits	*/
		stat = write_su_cm(2, 2, afe->su_cm_2_2);	/* Set in PPU		*/
		/*
		 * Find gain, and set afe-gain accordingly
		 */
		afe->gain = (float)gv[dr].gain/10;			/* Set actual gain	*/
		if (alc->debug)
			fprintf(debug_channel,"density_range: Real gain = %f\n", afe->gain);
	} else {
		stat = tomip1("de", dr & EXIST_GAIN);
	}
	return(stat);
} /* end density_range */

/*
 * F I N D _ I N D E X -- Find index in gv, corresponding to given gain value
 */
static int
find_index(gain)
float gain;
{
	int i, nranges;
	nranges = sizeof(gv)/sizeof(struct gain_range);

	for (i=0; i < nranges; i++)
		if (fabs(gain - (float)gv[i].gain/10) < 0.01) break;
	if (i == nranges) i = 0;	/* Return zero index */
	if (alc->debug)
		fprintf(debug_channel, "find_index: gain = %f, index = %d\n", gain, i);
	return(i);
} /* end find_index */

/*
 * W R I T E _ A F E -- Write all afe->su_cm fields to miproc
 */
static int
write_afe()
{
	write_su_cm(0, 3, afe->su_cm_0_3);
	write_su_cm(2, 2, afe->su_cm_2_2);
	write_su_cm(2, 3, afe->su_cm_2_3);
	write_su_cm(2, 6, afe->su_cm_2_6);
	write_su_cm(2, 7, afe->su_cm_2_7);
	return(0);
}

/*
 *	W R I T E _ S U _ C M -- Write value 'val' to sub_unit 'su' command 'cm'
 */
static int
write_su_cm(su, cm, val)
int su, cm, val;
{
	su &= 0xff; cm &= 0xff; val &= 0xff;
	if (alc->debug)
		fprintf(debug_channel, "Write_su_cm: 0x%X to su 0x%X, cmd 0x%X. (wtm=%d)\n",
			val, su, cm, wtm);
	if (!wtm) return(0);					/* Don't write to miproc */
	if (tomip2("O", su, cm) != FS_OK)		/* Select o/p sub-unit & cmd */
		return(afe_errno = AFE_ME);			/* Miproc error	*/
	if (tomip1("W", val) != FS_OK)			/* Write value	*/
		return(afe_errno = AFE_ME);			/* Miproc error	*/
	return(0);
} /* end write_su_cm */

