/*
 *	P I A S P E C I A L . C -- Special functions for reading from pia
 *
 *  Written: D C Burgess
 *           Image Recognition Systems
 *           720 Birchwood Boulevard
 *           Birchwood Science Park
 *           Warrington
 *           Cheshire
 *           WA3 7PX
 *
 *	Copyright (c) and intellectual property rights Image Recognition Systems (1989)
 *
 *  Date:    28th March 1989
 *
 *	These functions are in addition to the piaopen etc. which (currently)
 *	form part of gplib.
 *	These programs run at user level and so disc transfers and terminal
 *	interrupts have higher priority.
 *
 *	Modifications
 *	-------------
 *	23 May 1989	dcb 	Remove debug msg.
 *	17 May 1989	dcb		Cope with blank pixel counts > 255
 */
#include <stdio.h>

struct MC68230 {			/* Structure definition for PIA */
	char dum0;
	unsigned char PGCR;
	char dum1;
	unsigned char PSRR;
	char dum2;
	unsigned char PADDR;
	char dum3;
	unsigned char PBDDR;
	char dum4;
	unsigned char PCDDR;
	char dum5;
	unsigned char PIVR;
	char dum6;
	unsigned char PACR;
	char dum7;
	unsigned char PBCR;
	char dum8;
	unsigned char PADR;
	char dum9;
	unsigned char PBDR;
	char duma;
	unsigned char PAAR;
	char dumb;
	unsigned char PBAR;
	char dumc;
	unsigned char PCDR;
	char dumd;
	unsigned char PSR;
	unsigned short skip[2];
	unsigned short TIMER[16];
};

/* register bit definitions.  Only those required are specified */

/* PGCR */
#define MODE1			0100
#define H12_ENA			0020
#define H34_ENA			0040
#define HS_ASS_HIGH	0017
#define HS_ASS_LOW	0000
#define H1_ASSERT_HIGH		0001
#define H2_ASSERT_HIGH		0002
#define H3_ASSERT_HIGH		0004
#define H4_ASSERT_HIGH		0010

/* PSRR is cleared to zero by RESET */

/* P[AB]DDR */
#define DDREAD			0000
#define DDWRITE			0377

/* PCDDR */
#define CDDSET			0x13	/* enables output on 0,1,4 */

/* PCDR */
#define CDREAD			0xef	/* enable,  set bidirectional data drivers */
#define CDWRITE			0xec	/* ditto for writing */

/* P[AB]CR */
#define INPUT			0000
#define OUTPUT			0100
#define H24_STAT_ONLY		0000
#define H24_NEG_ALWAYS		0040
#define H24_ASSERT_ALWAYS	0050
#define H24_PULSE_HS		0070
#define H24_INTLOC_HS		0060
#define H13_STAT_CTRL		0001

/* PSR */
/* The instantaneous levels are 0 if logic is low and 1 if logic is high */
#define H1_LEVEL		0020
#define H2-LEVEL		0040
#define H3_LEVEL		0100
#define H4_LEVEL		0200
#define H1S			0001
#define H2S			0002
#define H3S			0004
#define H4S			0010

/* direction definitions as used in Unix system programs */
#define BREAD	0
#define BWRITE	1

static struct piastatus {
	int open;
	int reading;
	int writing;
};

static int pia_eof;
static struct piastatus piastatus = {0};

/*
 *	R E A D P I A 8 -- read a buffer from the pia.
 *	Input:	b = buffer to read to
 *			l = Data items to read
 *	Output:	Return status = number of data items actually read
 */
readpia8(b,l)
register unsigned char *b;
register int l;
{
	/*
	 * check if allowed to read, error return -1 otherwise.
	 * also check for positive, even requested byte count.
	 */
	/*
	 *	Can't check if allowed to read, this status is not available to us
	 *	Need extra functions!!!
	 */
#ifdef CANTCHECK
	if (piastatus.open != BREAD+1 || piastatus.reading != 0)
		return(-1);
#endif
	/*fprintf(stderr, "readpia8: Buf=0x%X, request 0x%x bytes\n", b, l);*/
	if (l <= 0 || (l&1) != 0)	/* l must be +ve, even */
		return(-1);
	piastatus.reading = 1;

	/*
	 * Now loop waiting for data.  The handshake is automatically
	 * handled by the 68230 when data is read from PBDR.  PADR
	 * must be read first.  Note that the 68000 "movep" insruction
	 * would is best here.
	 */
	return(subr8pia(b,l));
}

#asm
*
*	S U B R 8 P I A - Read from pia and store as 8 bit values
*
*	subr8pia -	Read data from pia, and compress to 8 bit values using the
*				following rules.
*	1. Neither B15, nor B14 set, store data value (Normal)
*	2. Blank pixel count (B15) set, store as zero then count (2 bytes)
*	3. Eol bit (b14) set, store as double zero (also 2 bytes).
*	NOTE
*	The EOLFLAG bit can occure in conjunction with a data value, or a blank
*	pixel count. Thus we need to check for both conditions. The max. data store
*	used could thus be four bytes (zero, blank count, zero, zero)
*
*	The handshake works in the following manner (I think)
*	1. H1_LEVEL indicates ready for data transfer
*	2. H3S is a data strobe
*	i.e.
*	H1_LEVEL | H3S | Action
*   ------------------------------------------------
*      0     |  0  | Return - No more data available
*      0     |  1  | Read last of data (double buffered) while H3S set
*      1     |  0  | Maybe missed data strb, Try again, to TRY_COUNT then sleep
*      1     |  1  | Normal - read data
*
*	Register usage:
*	a2 - Current buffer position for data to go
*	a3 - End of buffer position - must not be equalled or exceeded
*	a4 - Point at psr (status) pia register
*	a5 - Point at padr (pia A data register)
*	d0 - General purpose & return status
*	d1 - H1_LEVEL, and H3S mask
*	d2 - Original start buffer position
*	d3 - Repeat count before tsleep in read
*	d4 - General purpose
*	d5 - Top byte mask
*
H1_LEVEL	equ	$10				| Mask for bit 4
H3S			equ	$04				| Mask for bit 2
B_H1_LEVEL	equ	4				| Bit 4
B_H3S		equ	2				| Bit 2
BLANK_PIX	equ	15				| Bit 15 indicates blank pixel count
EOL_BIT		equ	14				| Bit 14 indicates end of line
BLANK		equ	$8000			| Bit 15 = blank pixel count
EOLFLAG		equ	$4000			| Bit 14 = End of line flag
TRY_COUNT	equ	500				| Count before sleeping
pia_base	equ	$fe0800			| Fixed base for pia address
padr		equ	$11				| Port A data register (Base reg. for all data)
psr			equ	$1b				| Port status register
*
*	S U B R 8 P I A
*
subr8pia:
	movem.l	d1-d5/a2-a5,-(sp)
	movea.l	d0,a2				| Buffer pointer
	move.l	d0,d2				| Keep a record in d2 also
	move.l	d1,d4				| Data items to read (Less blank counts + eols)
	moveq.l	#-72,d0				| Stack space we are about to use (2*9longs)
	bsr		_stkcheck			| Check we don't exceed available space
	move.l	a2,d0				| Start of buffer
	add.l	d4,d0				| Add required reads
	subq.l	#1,d0				| Less 1 ensures no overrun on blanks, or eols
	subq.l	#2,d0				| Leave space for eol sequence
	movea.l	d0,a3				| End of buffer pointer
	move.l	#H1_LEVEL!H3S,d1	| Mask for handshake bits
	clr.l	d3					| Timeout count
	move.w	#$ff00,d5			| Top byte mask includes Blank pix ct, EOL flags
	lea		pia_base,a4			| Load up base of pia registers
	lea		padr(a4),a5			| Data register
	lea		psr(a4),a4			| Status register
*
*	This is the top of the main read loop
*
r8pia_rpt
	move.b	(a4),d0				| Get h/s status (psr)
	and.b	d1,d0				| Mask H1_LEVEL, and H3S bits
	cmp.b	d1,d0				| Both present ?
	bne.s	r8pia_nahs			| No - so see what other action to take
	movep.w	(a5),d0				| Get (16 bit) data (padr)
	move.b	d0,(a2)+			| Store data + incr ptr
	and.w	d5,d0				| See if any of Blank pix. or eol flags set
	bne.s	r8pia_bleol			| Yes, so see which + deal with accordingly
	cmpa.l	a2,a3				| Reached end of buffer ? (Can't write to (a3))
	bhi.s	r8pia_rpt			| No, so go for next data
	bra		r8pia_exit			| Yes, so exit
*
*	Test top byte of data for control bits.
*
r8pia_bleol
	clr.l	d3					| Convenient place to reset timer
	btst.l	#BLANK_PIX,d0		| Was it a blank pixel count ??
	beq.s	r8pia_eol			| No, EOL flag ?
*
*	Store blank pixel count as zero then count, or multiple
*	occurrances of zero + pixel count.
*
	subq.l	#1,a2				| Back-up
	move.b	(a2),d0				| Restore full data (Bottom byte)
	move.w	d0,d4				| Keep copy for later check of EOL flag
	andi.w	#^(BLANK!EOLFLAG),d0 | Remove flags to give just the blank pix count
r8pia_mb
	move.b	d0,d4				| Keep lower byte
	and.w	d5,d0				| Check no. of 256's
	beq.s	r8pia_be			| blank exit
	clr.b	(a2)+				| Zero indicates blank pix run
	move.b	#$ff,(a2)+			| Store count of 255
	cmpa.l	a2,a3				| Reached end of buffer ? (Can't write to (a3))
	bcs		r8pia_exit			| Yes, so exit
	move.b	d4,d0				| Restore lower byte
	sub.w	#255,d0				| Done another 255
	bgt.s	r8pia_mb			| More blank count to store
r8pia_be
	clr.b	(a2)+				| Zero indicates blank pix run
	move.b	d4,(a2)+			| Store remainder (Now in lower byte of d4)
	cmpa.l	a2,a3				| Reached end of buffer ? (Can't write to (a3))
	bcs		r8pia_exit			| Yes, so exit
	move.w	d4,d0				| All done, restore flags to d0 for EOL check
*
*	Store end of line as double zero
*
r8pia_eol
	btst.l	#EOL_BIT,d0			| Was it end of line?
	beq.s	r8pia_rpt			| No, go and process more data

	clr.b	(a2)+				| First zero = special
	clr.b	(a2)+				| second zero means end of line
	cmpa.l	a2,a3				| Reached end of buffer ? (Can't write to (a3))
	bhi.s	r8pia_rpt			| No, so go for next data
	bra		r8pia_exit			| Yes, so exit
*
*	Test handshake bits for other states
*
r8pia_nahs
	btst.l	#B_H1_LEVEL,d0		| Test H1_LEVEL
	beq.s	r8pia_h1nas			| Not asserted, check H3S as well
	btst.l	#B_H3S,d0			| Test H3S
	beq.s	r8pia_time			| H1_LEVEL asserted, H3S deasserted - try agn
*
*	H1_LEVEL not asserted, check H3S
*
r8pia_h1nas
	btst.l	#B_H3S,d0			| Test H3S
	beq.s	r8pia_exit			| Neither H1_LEVEL, nor H3S asserted, so leave
*
*	H1_LEVEL not asserted, H3S asserted - keep reading any double buffered data
*
	movep.w	(a5),d0				| Get (16 bit) data (padr)
	move.b	d0,(a2)+			| Store data + incr ptr
	and.w	d5,d0				| See if any of Blank pix. or eol flags set
	bne.s	r8pia_2bleol		| Yes, so see which + deal with accordingly
	cmpa.l	a2,a3				| Reached end of buffer ? (Can't write to (a3))
	bhi.s	r8pia_agn			| No, so go for next data
	bra.s	r8pia_exit			| Yes, so exit it really is the right exit point
r8pia_agn
	move.b	(a4),d0				| Get h/s status (psr)
	bra.s	r8pia_h1nas			| Now test & read again
*
*	Test top byte of data for control bits.
*
r8pia_2bleol
	btst.l	#BLANK_PIX,d0		| Was it a blank pixel count ??
	beq.s	r8pia_2eol			| No, must have been eol flag (we assume)
*
*	Store blank pixel count as zero then count, or multiple
*	occurrances of zero + pixel count.
*
	subq.l	#1,a2				| Back-up
	move.b	(a2),d0				| Restore full data (Bottom byte)
	move.w	d0,d4				| Keep copy for later check of EOL flag
	andi.w	#^(BLANK!EOLFLAG),d0 | Remove flags to give just the blank pix count
r8pia_2mb
	move.b	d0,d4				| Keep lower byte
	and.w	d5,d0				| Check no. of 256's
	beq.s	r8pia_2be			| blank exit
	clr.b	(a2)+				| Zero indicates blank pix run
	move.b	#$ff,(a2)+			| Store count of 255
	cmpa.l	a2,a3				| Reached end of buffer ? (Can't write to (a3))
	bcs		r8pia_exit			| Yes, so exit
	move.b	d4,d0				| Restore lower byte
	sub.w	#255,d0				| Done another 255
	bgt.s	r8pia_2mb			| More blank count to store
r8pia_2be
	clr.b	(a2)+				| Zero indicates blank pix run
	move.b	d4,(a2)+			| Store remainder (Now in lower byte of d4)
	cmpa.l	a2,a3				| Reached end of buffer ? (Can't write to (a3))
	bcs.s	r8pia_exit			| Yes, so exit
	move.w	d4,d0				| All done, restore flags to d0 for EOL check
*
*	Store end of line as double zero
*
r8pia_2eol
	btst.l	#EOL_BIT,d0			| Was it EOL ?
	beq.s	r8pia_agn			| No, go for more data

	clr.b	(a2)+				| First zero = special code
	clr.b	(a2)+				| second zero means end of line
	cmpa.l	a2,a3				| Reached end of buffer ? (Can't write to (a3))
	bhi.s	r8pia_agn			| No, so go for next data
	bra.s	r8pia_exit			| Yes, so exit
*
*	Incr. count, if it reaches TRY_COUNT, then
*	give other processes a chance by tsleeping
*
r8pia_time
	addq.l	#1,d3				| Incr. count
	cmpi.l	#TRY_COUNT,d3		| Reached TRY_COUNT ??
	bne		r8pia_rpt			| No, so test h/s again etc.
	move.l	d0,d3
	moveq.l	#1,d0				| Count for tsleep
	bsr		tsleep				| do it
	move.l	d3,d0
	clr.l	d3					| Zero count
	bra		r8pia_rpt			| Try again
*
*	Here on exit
*
r8pia_exit
	clr.b	(a2)+				| Ensure last line correctly terminated
	clr.b	(a2)+
	move.l	a2,d0				| Current pos
	sub.l	d2,d0				| Return bytes in buffer
	movem.l	(sp)+,a2-a5/d1-d5	| restore registers
	rts							| and return

#endasm
