/*
 *		M P F U N C S . C --
 *
 * Master process functions for simplified control of
 * multi-tasking and multi-processing for MRC M68000 - OS9
 * chromosome analysis system.
 *
 * These apply the same function to ALL active slaves (which are
 * detected by having kcont->processor[slave] non-zero).
 *
 * They all wait for completion before returning to the caller,
 * so if you want something to happen asynchronously you will have
 * to do it differently.
 *
 *           Image Recognition Systems
 *           720 Birchwood Boulevard
 *           Birchwood Science Park
 *           Warrington
 *           Cheshire
 *           WA3 7PX
 *
 *	Copyright (c) and intellectual property rights Image Recognition Systems (1988)
 *
 *  Taken from: Jim Piper et al.
 *				Medical Research Council
 *				Clinical and Population Cytogenetics Unit
 *				Edinburgh
 *
 *  Date:    1st November 1985.
 *
 *  Modifications
 *
 *	 1 Mar 1989	SEAJ		Added PROCESSED flag
 *  16 Dec 1988	SEAJ		Make sure sing_mp_k_serv copes with NULL objects returned
 *							from mpcopyobj.
 *	30 Jun 1988	SCG/SEAJ	mod in func_clear_bits; clear bits of ACTIVE
 *							objects only.
 *  22 Sep 1987	JIMP		Mods so that MASTCOPY not destroyed by multi-processor
 *							display routines.
 *  31 Mar 1987 JP			Changes to 30-03-87 stuff, plus some tidying.
 *  30 Mar 1987 JP			New single-object function for karyotype services.
 *  13 Mar 1987 JP			New function for karyotype services.
 */

#include <stdio.h>
#include <multip.h>
#include <wstruct.h>
#include <chromanal.h>
#include <kcontrol.h>


/*
 *	M P C L E A R  --  multi-slave screen clear
 *
 */
mpclear(kcont)
register struct kcontrol *kcont;
{
	register int slave;
	
	kcont->auxstat = 0;
	for (slave=0; slave<=NSLAVES; slave++)
		if (kcont->processor[slave] != 0)
		    commdslave(kcont,slave,SCREEN_CLEAR);
	while (kcont->auxstat != ALL_CLEAR)
		tsleep(2);
}


/*
 *	M P C P I C  ---
 *
 * multi-slave display of active objects in kcont->ool,
 * using frame kcont->f.
 */
mpcpic(kcont,colour,intens)
register struct kcontrol *kcont;
{
	/* set up colour and intensity */
	kcont->colour = colour;
	kcont->intens = intens;
	
	mpstandard(kcont,COM_DISPLAY_LIST);
}


/*
 *	M P I P I C  --
 *
 * multi-slave display of active objects in kcont->rol,
 * using appropriate frames from kcont->ivf.  Each slave
 * will only display an object marked as belonging to its
 * processor in the "MPROC" bits of the activity list.
 */
mpipic(kcont,colour,intens)
register struct kcontrol *kcont;
{
	/* set up colour and intensity */
	kcont->colour = colour;
	kcont->intens = intens;
	
	mpstandard(kcont,INDV_DISPLAY_LIST);
}

/*
 *	M P S T A N D A R D  ---  
 *
 *	standard multi-processor multi-object function driver
 */
mpstandard(kcont,funccode)
register struct kcontrol *kcont;
{
	register int ac,i,slave;

	/* clear function bits in activity list */
	func_clear_bits(kcont,(FUNCBEGIN|FUNCEND));

	/* clear function bits in slave status words, and start slaves */
	func_start_slaves(kcont,funccode);

	/* wait for slaves to finish */
	func_finish_slaves(kcont);
}

/*
 *	F U N C _ S T A R T _ S L A V E  --
 *
 * clear function bits in slave status words, and start slaves 
 */
func_start_slaves(kcont,funccode)
register struct kcontrol *kcont;
{
	register int slave,ac;
	for (slave=0; slave<=NSLAVES; slave++) {
		if (kcont->processor[slave] != 0) {
			kcont->status[slave] &= ~(FUNCBEGIN|FUNCEND|MAXNCHANGE);
			commdslave(kcont,slave,funccode);
		}
	}
}

/*
 *	F U N C _ F I N I S H _ S L A V E S  -- wait for slaves to finish 
 *
 */
func_finish_slaves(kcont)
register struct kcontrol *kcont;
{
	register int slave,ac;
	for (slave=0; slave<=NSLAVES; slave++) {
		if (kcont->processor[slave] != 0) {
			do {
				ac = kcont->status[slave];
				ac &= FUNCEND;
				if (ac == 0)
					tsleep(2);
			} while (ac == 0);
		}
	}
}

/*
 *	F U N C _ C L E A R _ B I T S  --
 *
 *	clear function bits in all activity words 
 */
func_clear_bits(kcont,bits)
register struct kcontrol *kcont;
unsigned short bits;
{
	register int i;
	for (i=0; i<kcont->maxnumber; i++) 
		if(kcont->acl[i] & ACTIVE)  kcont->acl[i] &= ~bits;
}

/*
 *	M P L I K L  --  multi-processor likelihood computation
 */
mplikl(kcont)
register struct kcontrol *kcont;
{
	register int ac,i,slave;

	/* clear function bits in slave status words, and start slaves */
	func_start_slaves(kcont,CHROM_LIKL_MEASURE);

	/* wait for slaves to finish */
	func_finish_slaves(kcont);
}

/*
 *	M P _ K _ S E R V  --
 *
 *	standard multi-processor multi-object karyotype service driver
 */
mp_k_serv(kcont,funccode)
register struct kcontrol *kcont;
{
	register int ac,i,slave,slavep,alldone;
	register SMALL *statpt;
	struct chromosome *mobj, *nobj, *mpcopyobj();

	/* clear function bits in activity list */
	func_clear_bits(kcont,(FUNCBEGIN|FUNCEND|MASTCOPY|MPROC|SLFREE));
	
	/* clear function bits in slave status words, and start slaves */
	func_start_slaves(kcont,funccode);

	/* loop copying back processed but uncopied objects */
	i = 0;
	alldone = 1;
	while (i<kcont->maxnumber) {
		statpt = kcont->acl + i;
		ac = *statpt;
		if ((ac & ACTIVE) && !(ac&FUNCEND))
			alldone = 0;
		if ((ac&ACTIVE) && (ac&FUNCEND) && !(ac&MASTCOPY)) {
			/* slave processor number */
			slavep = (ac & MPROC) >> 6;
			nobj = kcont->mol[i];
			kcont->mol[i] = mobj = mpcopyobj(convaddr(kcont->rol[i],slavep,MASTER),
								slavep,MASTER,1);
			/*
			 * mpcopyobj(,,,1) may make a NEW propertylist.
			 * Attach the original.  Free old mol[] object
			 * if it different to the kol[] (keeplist) version.
			 */
			if (mobj->plist != NULL)
				Free(mobj->plist);
			mobj->plist = nobj->plist;
			if (nobj != kcont->kol[i]) {
				nobj->plist = NULL;
				freeobj(nobj);
			}
			*statpt |= MASTCOPY;
		}
		i++;
		if (i == kcont->maxnumber && !alldone) {
			i = 0;
			alldone = 1;
			tsleep(5);
		}
	}

	/* wait for slaves to finish */
	func_finish_slaves(kcont);
}

/*
 *	S I N G _ M P _ K _ S E R V  --
 *
 * Standard multi-processor single-object karyotype service driver.
 * All relevant slaves are started, first available does the service.
 */
struct chromosome *sing_mp_k_serv(kcont,funccode,objnum)
register struct kcontrol *kcont;
{
	register int ac,i,slave,slavep;
	register SMALL *statpt;
	struct chromosome *mobj, *nobj, *mpcopyobj();

	/* set function bits in activity list */
	for (i=0; i<kcont->maxnumber; i++) 
		kcont->acl[i] |= (FUNCBEGIN|FUNCEND|MASTCOPY|MPROC|SLFREE|PROCESSED);

	/* clear function bits in activity word of specified object */
	kcont->acl[objnum] &= ~(FUNCBEGIN|FUNCEND|MASTCOPY|MPROC|SLFREE|PROCESSED);
	
	/* clear function bits in slave status words, and start slaves */
	func_start_slaves(kcont,funccode);

	/* wait for completion */
	statpt = kcont->acl + objnum;
	while (!(*statpt & FUNCEND))
		tsleep(2);
	/* slave processor number */
	slavep = (*statpt & MPROC) >> 6;
	nobj = kcont->mol[objnum];
	mobj = mpcopyobj(convaddr(kcont->rol[objnum], slavep,MASTER),
			slavep,MASTER,1);
	/*
	 * mpcopyobj(,,,1) may make a NEW propertylist.
	 * Attach the original.  Free old mol[] object
	 * if it different to the kol[] (keeplist) version.
	 */
	if (mobj != NULL) {
		if (mobj->plist != NULL)
			Free(mobj->plist);
		mobj->plist = nobj->plist;
	}
	*statpt |= MASTCOPY;

	/* wait for slaves to finish */
	func_finish_slaves(kcont);
	return(mobj);
}
