/*
 * mpcopyobj.c			Jim Piper			24 June 1985
 *
 * Copy object structure from one processor to another,
 * setting up a new hierarchy of structures in the local memory
 * of the second processor.
 * If flag "all" is set to zero, only copy those parts of
 * hierarchy which contain pointer variables.  If "all" is true,
 * also copy tables of grey-values, interval end points, polygon
 * vertices, etc..
 *
 * If processors "p" and "q" are the same, then the copy is performed
 * anyway, so this routine provides a general object duplicator.
 *
 *****************************************************************************
 *
 * Modifications
 *
 *  16 May 1991		ih		Check for NULL cvhdom parm in mpccvhdom
 *	30 Oct 1987		CAS		Cope with NULL object ptr
 * jimp	03-09-87	Linkcount setting removed from rectangular grey-table
 * CAS	05-03-87	Cope with type 11 grey tables.....
 * CAS	12-01-87	Return NULL from mpcopyobj rather than exit(1)...
 * JP	12-11-86	Includes 
 * JP	05-11-86	Extended for convex hull.
 * JP	05-11-86	Boundary copy function corrected.
 * JP	30-10-86	Extended for polygons and boundaries.
 * JP	30-10-86	Care taken with NULL structures.
 *****************************************************************************
 */

#include <stdio.h>
#include <wstruct.h>


struct object *mpcopyobj(obj,p,q,all)
struct object *obj;
{
	struct object *makemain();
	struct intervaldomain *mpcidom();
	struct polygondomain *mpcpolydom();
	struct valuetable *mpcvdom();
	struct propertylist *mpcplist();
	struct boundlist *mpcboundlist();
	struct cvhdom *mpccvhdom();

	if (obj == NULL)
		return(NULL);	
	switch (obj->type) {
		case 1:
			return( makemain(1,
				mpcidom(convaddr(obj->idom,p,q),p,q,all),
				mpcvdom(convaddr(obj->vdom,p,q),p,q,all),
				mpcplist(convaddr(obj->plist,p,q),p,q,all),
				obj));

		case 10:
			return( makemain(10,
				mpcpolydom(convaddr(obj->idom,p,q),p,q,all),
				NULL,
				mpcplist(convaddr(obj->plist,p,q),p,q,all),
				obj));

		case 11:
			return( makemain(11,
				mpcboundlist(convaddr(obj->idom,p,q),p,q,all),
				NULL,
				mpcplist(convaddr(obj->plist,p,q),p,q,all),
				obj));

		case 12:
			return( makemain(13,
				mpcpolydom(convaddr(obj->idom,p,q),p,q,all),
				mpccvhdom(convaddr(obj->vdom,p,q),p,q,all),
				mpcplist(convaddr(obj->plist,p,q),p,q,all),
				obj));

		default:
			fprintf(stderr,"MPCOPYOBJ: cannot copy type %d\n",obj->type);				
			return(0);
	}
}


struct intervaldomain *mpcidom(idom,p,q,all)
struct intervaldomain *idom;
{
	struct intervaldomain *ndom, *makedomain();
	struct intervalline *ivln, *jvln;
	struct interval *itvl, *jtvl, *ktvl;
	int i,l,intcount;

	if (idom == NULL)
		return(NULL);
	if (idom->type == 2 && all == 0)
		return(idom);
	ndom = makedomain(idom->type,idom->line1,idom->lastln,
			idom->kol1,idom->lastkl);
	if (ndom->type == 2)
		return(ndom);
	/* fill in the intervalline structures */
	ivln = ndom->intvlines;
	jvln = (struct intervalline *) convaddr(idom->intvlines,p,q);
	intcount = 0;
	for (i=ndom->line1; i<=ndom->lastln; i++) {
		ivln->nintvs = jvln->nintvs;
		ivln->intvs = (struct interval *) convaddr(jvln->intvs,p,q);
		intcount += ivln->nintvs;
		ivln++;
		jvln++;
	}
	/* interval table required ? Allocate space, copy, modify tables */
	if (all != 0) {
		itvl = (struct interval *) Malloc(intcount*sizeof(struct interval));
		ndom->freeptr = (char *) itvl;
		ivln = ndom->intvlines;
		for (l=ndom->line1; l<=ndom->lastln; l++) {
			jtvl = itvl;
			ktvl = ivln->intvs;
			for (i=0; i<ivln->nintvs; i++) {
				itvl->ileft = ktvl->ileft;
				itvl->iright = ktvl->iright;
				itvl++;
				ktvl++;
			}
			ivln->intvs = jtvl;
			ivln++;
		}
	}
	return(ndom);
}


struct polygondomain *mpcpolydom(pdom,p,q,all)
struct polygondomain *pdom;
{
	struct polygondomain *makepolydmn();
	if (pdom == NULL)
		return(NULL);
	return(makepolydmn(pdom->type, convaddr(pdom->vtx,p,q),
		pdom->nvertices,pdom->maxvertices,all));
}


struct valuetable *mpcvdom(vdom,p,q,all)
struct valuetable *vdom;
{
	struct valuetable *wdom, *makevaluetb();
	struct rectvaltb *rdom1, *rdom2, *makerectvaltb();
	struct valueline *vln,*wln,**vtbl,**wtbl;
	int i,l,gsize;
	GREY *wg,*vg,*gwas;
	
	if (vdom == NULL)
		return(NULL);
	switch(vdom->type) {
		case 1:
			wdom = makevaluetb(1,vdom->line1,vdom->lastln,vdom->bckgrnd,NULL);
			/* set up pointers and compute size */
			gsize = 0;
			vtbl = (struct valueline **) convaddr(vdom->vtblines,p,q);
			wtbl = wdom->vtblines;
			for (l=wdom->line1; l<=wdom->lastln; l++) {
				vln = (struct valueline *) convaddr(*vtbl,p,q);
				wln = *wtbl;
				wln->vkol1 = vln->vkol1;
				wln->vlastkl = vln->vlastkl;
				wln->values = (GREY *) convaddr(vln->values,p,q);
				gsize += (wln->vlastkl - wln->vkol1 + 1);
				vtbl++;
				wtbl++;
			}
			/* copy grey-table if required */
			if (all != 0) {
				wg = (GREY *) Malloc(gsize * sizeof(GREY));
				wdom->freeptr = (char *) wg;
				wtbl = wdom->vtblines;
				for (l=wdom->line1; l<=wdom->lastln; l++) {
					wln = *wtbl;
					gwas = wg;
					vg = wln->values;
					for (i=wln->vkol1; i<=wln->vlastkl; i++)
						*wg++ = *vg++;
					wln->values = gwas;
					wtbl++;
				}
			}
			return(wdom);

		case 11:
			rdom1 = (struct rectvaltb *) vdom;
			wg = (GREY *) convaddr(rdom1->values,p,q);
			rdom2 = makerectvaltb(11,rdom1->line1,rdom1->lastln,
				rdom1->kol1,rdom1->width,rdom1->bckgrnd,wg);
			/* copy grey-table if required */
			if (all != 0) {
				gsize = rdom2->width * (rdom2->lastln - rdom2->line1 + 1);
				vg = (GREY *) Malloc(gsize * sizeof(GREY));
				rdom2->values = (GREY *) vg;
				rdom2->freeptr = (char *) vg;
				for (i=1; i <= gsize; i++) {
					*vg++ = *wg++;
				}
			}
			return(rdom2);

		default:
			fprintf(stderr,"MPCVDOM: cannot mp-copy type %d\n",vdom->type);
			exit(1);
	}
}


struct propertylist *mpcplist(pl,p,q,all)
register char *pl;
{
	register char *ql;
	register int i,size;
	if (pl == NULL)
		return(NULL);
	if (all == 0)
		return((struct propertylist *) pl);
	else {
		size = ((struct propertylist *)pl)->size;
		ql = (char *) Malloc(size);
		bcopy(pl,ql,size);
		return((struct propertylist *) ql);
	}
}

struct boundlist *mpcboundlist(bdom,p,q,all)
struct boundlist *bdom;
{
	struct boundlist *blist;
	struct polygondomain *mpcpolydom();
	if (bdom == NULL)
		return(NULL);
	blist = (struct boundlist *) Malloc (sizeof(struct boundlist));
	blist->type = bdom->type;
	blist->up = NULL;
	blist->next = mpcboundlist(convaddr(bdom->next,p,q),p,q,all);
	blist->down = mpcboundlist(convaddr(bdom->down,p,q),p,q,all);
	blist->wrap = bdom->wrap;
	blist->poly = mpcpolydom(convaddr(bdom->poly,p,q),p,q,all);
	return(blist);
}


struct cvhdom *mpccvhdom(cvhd,p,q,all)
struct cvhdom *cvhd;
{
	struct cvhdom *c;
	int size;
	
	if (cvhd==NULL) return(NULL);
	size = all? cvhd->nchords*sizeof(struct chord): 0;
	c = (struct cvhdom *) Malloc(sizeof(struct cvhdom) + size);
	bcopy(cvhd,c,sizeof(struct cvhdom));
	if (all == 0)
		c->ch = (struct chord *) convaddr(cvhd->ch,p,q);
	else {
		c->ch = (struct chord *) (c+1);
		bcopy(convaddr(cvhd->ch,p,q),c->ch,size);
	}
}

static bcopy(a,b,c)
register char *a, *b;
register int c;
{
	while (c-- > 0)
		*b++ = *a++;
}
