/*
 *  S L _ F M E A S . C -- slave service function to compute features of 
 *						   chromosome objects (was kmeas.c)
 *
 *           Image Recognition Systems
 *           720 Birchwood Boulevard
 *           Birchwood Science Park
 *           Warrington
 *           Cheshire
 *           WA3 7PX
 *
 *	Copyright (c) and intellectual property rights Image Recognition Systems (1988)
 *
 *  Taken from: Jim Piper et al.
 *				Medical Research Council
 *				Clinical and Population Cytogenetics Unit
 *				Edinburgh
 *
 *  Date:    17th April 1985
 *
 *  Modifications
 *
 *  30 May 1991		ih		privatise dstring to avoid clash with sl_k_serv
 *	15 Jan 1991		CAS/GJP	Remove blobs (cond TRAVELLING_BLOBS) while measuring
 *	22 Mar 1990		dcb		Speed up main loop in kmeas, tsleep 20
 *	20 Mar 1990		GJP		Move slave numbers up
 *	15 Mar 1990		dcb		free_left returns TRUE if memory size passed as
 *							param can be allocated., else FALSE
 *							tsleep time increased to 10 in feature measure loop
 *   1 Mar 1989		SEAJ	Added activity bit PROCESSED.
 *	 6 Dec 1988		CAS		Remove messages into DEBUG defines
 *	 2 Nov 1987		SCG/CAS	Fixed stupid bug in wormcast mod... PRATT
 *	22 Oct 1987		CAS		Only wormcast in slave 1.. - Not ideal but to in
 *							practice will be OK
 *							Changed SEGEND processing to ensure no stragler
 *							objects left unmeasured
 * 24-09-87			jp/bdp	Get the numbers at bottom right for ALL displays.
 * 22-09-87			jimp	Modification to 18-09-87 stuff: don't free after
 *							MASTCOPY bit set, since MPROC no longer guaranteed
 *							(there could otherwise be problems in karyotype
 *							re-segmentation).  MASTCOPY is permanently set
 *							during karyotyping except when karyotype server
 *							functions are happening.
 * 18-09-87	jimp		Detect previously measured objects owned by this
 *			processor (NB should be OWNED BY THIS PROCESS, but that
 *			information not available in acl[] - does not matter
 *			so long as only one slave per processor).  Free the
 *			unwanted object and fix up acl[] appropriately.
 *			To make this more secure, always copy the input object,
 *			even if running in the master processor.
 * jimp	05-08-87	Leave the master to do the attention-drawing.
 *	29 Jan 1987		CAS			Cockup with kcont->FIP
 * JP	15-12-86	Wormcast tag changed to reflect module name change.
 * JP	14-11-86	Even if an operator disregards the "probotype" typing
 *			and calls a large object a chromosome (possibly owing
 *			to a finger fault), if it is too small or too big we
 *			only do the "basic" feature measurements.
 * JP	13-11-86	Reconstructed from divergent MRC and IRS versions.
 * JP	06-11-86	Name changed.
 * JP	31-10-86	If object type believed not CHROMOSOME, draw operator's
 *			further attention to this object (and don't measure).
 *			This replaces the previous test based on "maxdiam()".
 * JP	17-10-86	First guess at object type
 * JP	17-10-86	INTERACTING flag checked before running measurement
 *	 7 Oct 1986		CAS			Add ABORT option
 *  Get the size of remaining memory and shut up shop 
 * if insufficient left.    19/5/86 GJP
 */

#include <stdio.h>
#include <multip.h>
#include <wstruct.h>
#include <chromanal.h>
#include <kcontrol.h>
#include <display.h>
#include <fileheaders.h>

#define BASIC 1
#define ENOUGH_MEM 30000	/* single chunk */
#define MINDIAM 10
#define MINDENSITY 10
#define MAXAREA 5000
#define TRUE (1)
#define FALSE (0)

char astring[80];
#ifdef DEBUG
static char dstring[50];
#endif DEBUG

extern double xscale,yscale;

kmeas (slave, kcont)
struct kcontrol *kcont;
{
	struct object *obj, *robj;
	struct object *chromfeatures(), *mpcopyobj();
	struct object **ool, **rol;
	struct chromplist *plist;
	struct ipoint blob, **bol, *mblob;
	struct pframe *f;
	int slavep, SEGBEG, slavey, ddgstype, num_xpos, *kmstatus;
	SMALL *acl;
	register int i, segend, measuring;
	register SMALL *statpt, *kmaxnumber;
	register char *klock;
	int loopcnt=0;

#ifdef DEBUG
	int j;
#endif DEBUG

	/* slave process number must be converted to true processor number */
	slavep = kcont->processor[slave] >> P_OR;
	tsize(50,50);
	slavey = 112 + 80 * slave;
/*	slavey = 3800 + 80 * slave; GJP 13/5/86  */
#ifdef DEBUG
	sprintf(dstring,"slave %d pr_or %d pr_ss %d object",slave,slavep,
		kcont->processor[slave] & P_SS);
	dispstring(dstring,2400,slavey);
#endif DEBUG
	kcont->status[slave] |= MEASBEGIN;
	ool = (struct object **)convaddr(kcont->ool,MASTER,slavep);
	acl = (SMALL *)convaddr(kcont->acl,MASTER,slavep);
	rol = (struct object **)convaddr(kcont->rol,MASTER,slavep);
	bol = (struct ipoint **)convaddr(kcont->bol,MASTER,slavep);
	f = (struct pframe *)convaddr(kcont->f,MASTER,slavep);
	ddgstype = ((DDGS *)convaddr(kcont->ddgschannel,MASTER,slavep))->type;
	if (ddgstype == FS_VIRTUOSO || ddgstype == FS_VIRTUOSO_INVERT)
		num_xpos = 5940;
	else
		num_xpos = 3960;
	switch (kcont->FIP) {
		default:
		case DIGTV:
			xscale = yscale = 1.0;
			break;
		case DIGFIP:
			xscale = 1.0;
			yscale = 0.75;
			break;
		case DIGFIPEXP:
			xscale = 1.33;
			yscale = 1.0;
			break;
	}

	/* set up blob display */
	blob.type = 40;
	blob.style = 4;	/* fixed size hollow square */

	/*
	 * wait until at least one object present
	 */
	while (kcont->maxnumber <= 0)
		tsleep(10);

	/*
	 * Measure features and display in bottom corner of screen.
	 * Loop in range 1 - maxnumber until all active objects processed
 	 * Now wait, and scan list again so as to catch any composites etc.
 	 * subsequently re-designated as chromosomes.
 	 * Continue until either ABORT or SEGEND is flagged in the status word.
 	 * When SEGEND is flagged, do one last pass through the objects
	 *
	 * Don't measure features on tiny objects.  These will be noise,
	 * pieces or fragments and will either be junked, joined or
	 * treated separately in some fashion.  If left at end of segmentation
	 * then at least they have a property list (even if empty)
	 *
	 * Before measuring, check that we have enough free memory to do the job
	 */
#ifdef DEBUG
/* fprintf(stderr,"Memory at start %d\n",free_mem());*/
#endif DEBUG
	klock = &(kcont->lock);
	kmaxnumber = &(kcont->maxnumber);
	kmstatus = &(kcont->mstatus);
	measuring = TRUE;
	segend = FALSE;

	while (measuring) {
#ifdef DEBUG
		intens(0);
		dispstring(dstring,600,slavey);
		sprintf(dstring,"loopcount %d:6 mstatus %04o maxn %d status %04o",loopcnt++,
			kcont->mstatus,kcont->maxnumber,kcont->status[slave]);
		intens(1);
		dispstring(dstring,600,slavey);
		if(slavep==MASTER){
			intens(0);
			dispstring(astring,600,3700);
			for (j=0;j<kcont->maxnumber;j++){
				if(acl[j]&ACTIVE==ACTIVE)
					astring[j]='*';
				else
					astring[j]=' ';
			}
			intens(1);
			dispstring(astring,600,3700);
		};
#endif DEBUG
		statpt = acl;
		for (i=0; i < *kmaxnumber; i++, statpt++) {
/**********************************/
			lock7(klock);
			if (((*statpt & (ACTIVE|MSTART|INTERACTING)) == ACTIVE) &&
				free_left(ENOUGH_MEM)) {
				*statpt |= MSTART;			/* Let's do it ! */
				unlock7(klock);
/**********************************/
				if (slavep == 1) wormcast(kcont,slavep,"sl_fmeas");
				colour(OVERLAY1); intens(1);
				dispnum(i+1,num_xpos,slavey);
				obj = mpcopyobj(convaddr(ool[i],MASTER,slavep),MASTER,slavep,0);
#ifdef TRAVELLING_BLOBS
				/*
				 * Don't display if segmentation not yet started,
				 * e.g. while objects still being read from disc
				 */
				SEGBEG = kcont->mstatus&SEGBEGIN;
				if (SEGBEG != 0) {
					mblob = (struct ipoint *)convaddr(bol[i],MASTER,slavep);
					blob.k = mblob->k;
					blob.l = mblob->l;
					picframe(&blob,f);
				}
#endif
				/* type already known not chrom ?? */
				if (((struct chromosome *)obj)->plist->otype > 1)	
					robj = obj;
				else if (probotype(obj) > 1) {	/* believed not chromosome */
					robj = obj;
				}
				else {							/* believed to be chromosome */
					/*
					 * small object test - since operator may have forced
					 * chromosome classification on us.
					 */
					plist = (struct chromplist *) obj->plist;
					if (plist->area == 0)
						plist->area = area(obj);
					if (plist->mass == 0)
						plist->mass = mass(obj);
					if (maxdiam(obj) < MINDIAM ||
						plist->mass/plist->area < MINDENSITY ||
						plist->area > MAXAREA) {
							robj = obj;
							basicfeatures(robj);
					} else
						robj = chromfeatures(obj,BASIC);
				}
				colour(OVERLAY1);
				intens(0);
				dispnum(i+1,num_xpos,slavey);
#ifdef TRAVELLING_BLOBS
				if (SEGBEG != 0) {
					picframe(&blob,f);
				}
#endif
				intens(1);
				/*
				 * Can delete "obj" now.
				 * First set obj->plist to NULL since "obj" and "robj"
				 * share the property list.  Also test that it really is a
				 * new object (on occasion we may not have rotated original)
				 */
				if (obj != robj) {
					obj->plist = NULL;
					freeobj(obj);
				}
				/*
				 * Check see if any previously measured objects can be ditched
				 */
				check_free(kcont,slavep);
/**********************************/
				lock7(&(kcont->lock));
				rol[i] = robj;
				*statpt |= (MEND | (slavep << 6) | PROCESSED);
				unlock7(&(kcont->lock));
/**********************************/
			} /* end if ACTIVE */
			unlock7(klock);
/**********************************/
			if (kcont->mstatus & ABORT) {
				measuring = FALSE;
				break;
			}
		} /* end for i... */
		if (segend) measuring = FALSE;
		else if (kcont->mstatus & SEGEND) segend = TRUE; /* One last pass */
		check_free(kcont,slavep);/* Can we ditch prev. measured objs */
		tsleep(20);
	} /* end while measuring */

	kcont->status[slave] |= MEASEND;
#ifdef DEBUG
/*	if (free_mem()<ENOUGH_MEM)
		fprintf(stderr,"Kmeas-insufficient memory\n");*/
#endif DEBUG
}

/*
 *	F R E E _ L E F T --	Return TRUE if memory 'size' can be allocated.,
 *							else FALSE
 */
static int
free_left(size)
char *size;
{
	char *mptr, *Malloc();
	int rval;

	Alloc_wait(0);	/* Don't wait for memory */
	if ((mptr = Malloc(size)) == NULL)
		rval = FALSE;
	else {
		Free(mptr);
		rval = TRUE;
	}
	Alloc_wait(1);	/* Wait again */
	return(rval);
}

check_free(kcont,slavep)
register struct kcontrol *kcont;
{
	SMALL *acl;
	register SMALL *statpt;
	register char *klock;
	register struct object **rol;
	register int i;

	acl = (SMALL *)convaddr(kcont->acl,MASTER,slavep);
	rol = (struct object **)convaddr(kcont->rol,MASTER,slavep);
	statpt = acl;
	klock = &(kcont->lock);
	for (i=0; i<kcont->maxnumber; i++, statpt++) {
		lock7(klock);
		if ((*statpt & (ACTIVE|MSTART|MEND|INTERACTING|MASTCOPY)) ==
		    (MSTART|MEND) && ((*statpt&MPROC) >> 6) == slavep) {
			unlock7(klock);
			rol[i]->plist = NULL;
			freeobj(rol[i]);
			lock7(klock);
			rol[i] = NULL;
			*statpt &= ~(MSTART|MEND|MPROC);
#ifdef DEBUG
/*			fprintf(stderr,"Slave freeing rotated object number %d (mem %d)\n",i+1,free_mem());*/
#endif DEBUG
		}
		unlock7(klock);
	}
}
