/*
 * slavecontrol.c		Jim Piper		1/11/85
 *
 * Master process functions for controlling the master-slave
 * multi-tasking and multi-processing for MRC M68000 - OS9
 * chromosome analysis system.
 *
 * Modifications:
 *
 *  25 Mar 1992 jimp	Now forks "kslave" - BEWARE
 *  31 May 1991		ih			Tidy up slave death checks - also tsleep(2) - 10
 *	10 Apr 1991		ih			Add a check for a dead slave in master
 *	22 Feb 1991		CAS			Once and for all, make the fork %d bigenough
 *	11 Jan 1991		CAS			Options to get slave ack for commands
 *	23 Oct 1989		dcb			Pass initslave the slave priority
 *	14 Jul 1988		CAS			Pass slave the addr map table
 *	29 Jan 1987		CAS			Cockup with kcont->FIP
 *	30 Sep 1986		CAS			Removed 'exit()' calls from here
 *								and return usefull status instead
 *								Also hacked out interrupt() to its own
 *								module
 *	22 Apr 1986		CAS + GJP	Modified initslave to work with CAS's
 *								chromosome independant multiprocessor bits.
 */

#include <stdio.h>
#include <module.h>
#include <procid.h>
#include <multip.h>
#include <wstruct.h>
#include <kcontrol.h>
#include <fileheaders.h>

#define COMMAND_AFFIRM		/* Make sure slave has seen and ack'd commands */
#define SLAVE_DEATH "My poor slave hath died\n"

static struct kcontrol *kcont;


/*
 * set up a kcontrol structure with default values
 */
initkcontrol(kc)
register struct kcontrol *kc;
{
	register int slave;

	kcont = kc;	
	kc->seat = 0;
	kc->terminal = "/term";
	kc->maxnumber = 0;
	kc->FIP = DIGFIP;
	unlock7(&(kc->lock));
	kc->mstatus = 0;
	for (slave=0; slave<=NSLAVES; slave++) {
		kc->command[slave] = 0;
		kc->status[slave] = 0;
		kc->processor[slave] = 0;
	}
}


/*
 * fork a slave process in a specified processor
 * (using facilities of "slaveshell" in slave processors)
 *	returns:
 *		1  - if activated ok
 *		0  - if already active
 *		-1 - if illegal slave number or fork in slave failed
 */
initslave(kc,slave,processor, prior)
struct kcontrol *kc;
{
	int *convaddr(), pid;
	char params[20];
	if (slave < 0 || slave >= NSLAVES) {
		fprintf(stderr,"INITSLAVE: illegal slave %d\n",slave);
		return(-1);
	}
	if (kc->processor[slave] != 0) {
		return(0);
	}
	kc->processor[slave] = processor<<P_OR;
/*
 * Send address config map to slave processor in case this is first time used.
 * This call should eventually be moved and done as part of the system startup
 * configuration together with the bootp.  It is done here since kmr currently
 * sorts out the slave config.  There is a potential problem re-writing
 * the config table but since there are no async events in the slave using
 * the table and it is not actually being changed (after the initial set up)
 * then it probably doesn't matter that much
 */
	slavesetmap(processor);
	sprintf(params,"%d %d\n",(int) convaddr(kc,MASTER,processor),slave);
	pid = slavefork(processor,"kslave",strlen(params),params,0,0,0,prior,3);
	if (pid) {
		kc->processor[slave] |= pid;
	} else {
		fprintf(stderr,"SLAVE %d: fork in processor %d failed\n",slave,processor);
		kc->processor[slave] = 0;
		return(-1);
	}
	return(1);
}


/*
 *	K I L L S L A V E  --  send a signal to a specified slave process
 * (using facilities of "slaveshell" in slave processors)
 *
 *	returns:
 *		1  - if killed ok
 *		0  - if not active (already killed)
 *		-1 - if illegal slave number
 */
killslave(kc,slave,sig)
struct kcontrol *kc;
{
	int processor, *convaddr();
	char params[20];
	if (slave < 0 || slave >= NSLAVES) {
		fprintf(stderr,"KILLSLAVE: illegal slave %d\n",slave);
		return(-1);
	}
	if (kc->processor[slave] == 0) {
		return(0);
	}
	processor = kc->processor[slave] >> P_OR;
	if (slavekill(processor,kc->processor[slave] & P_SS, sig) != -1) {
		if (sig == 0 || sig == 2 || sig == 3)
			kc->processor[slave] = 0;
	}
	return(1);
}


/*
 *	W A I T S L A V E  --  wait for a particular slave to report a 
 *							particular status
 *
 *	Returns:
 *		1  - if ok
 *		-1 - if illegal slave
 */
waitslave(kc,slave,statusbits)
struct kcontrol *kc;
{
	register int stat,sid;
	int wait;

	sid=kc->processor[slave]&P_SS;
	if (!sid) {
#ifdef DEBUG
		fprintf(stderr,"Waitslave: slave dead!!!\n");
#endif DEBUG
		return(-1);
	}
	if (slave < 0 || slave >= NSLAVES) {
		fprintf(stderr,"WAITSLAVE: illegal slave %d\n",slave);
		return(-1);
	}
	/* FRIG ROUND OS9 V1.2 compiler BUG */
	do {
		stat = kc->status[slave];
		stat &= statusbits;
		if (stat != statusbits) {
			if (slave==MASTER) {
				wait=Alloc_wait(1);
				while (waitchild());
				if (testproc(sid)==-1) {
#ifdef DEBUG
					char buffer[80];
					sprintf(buffer,"Waitslave %s",SLAVE_DEATH);
					fprintf(stderr,"%s",buffer);
#endif DEBUG
					kc->processor[slave]=0;
					Alloc_wait(wait);
					return(-1);
				}
				Alloc_wait(wait);
			}
			tsleep(10);
		}
#ifdef DEBUG
		/* DEBUG - buzz bell if lock is locked !! */
		if (kc->lock != 0)
			putc('\007',stderr);
#endif DEBUG
	} while (stat != statusbits);
	return(1);
}


/*
 * master and slave status
 */
slavestat(kc)
struct kcontrol *kc;
{
	register int slave;
	fprintf(stderr,"master status %06o maxnumber %d\n",
		kc->mstatus,kc->maxnumber);
	for (slave = 0; slave < NSLAVES; slave++) {
		fprintf(stderr,"slave %d ",slave);
		if (kc->processor[slave] == 0)
			fprintf(stderr,"not active\n");
		else {
			fprintf(stderr,"in proc %d ",kc->processor[slave] >> P_OR);
			fprintf(stderr,"func %d ",kc->command[slave]);
			fprintf(stderr,"status %06o\n",kc->status[slave]);
		}
	}
}


/*
 *	C O M M D S L A V E  --  send a command to a specified slave
 *
 *	Returns:
 *		1  - if ok
 *		-1 - if illegal slave
 */
commdslave(kc,slave,command)
struct kcontrol *kc;
{
	int sid=kc->processor[MASTER]&P_SS;
	int wait;

	if (slave < 0 || slave >= NSLAVES) {
		fprintf(stderr,"COMMDSLAVE: illegal slave %d\n",slave);
		return(-1);
	}
	kc->command[slave] = command;
#ifdef COMMAND_AFFIRM
	while (kc->command[slave]) {
		if (slave==MASTER) {
			wait=Alloc_wait(1);
			while (waitchild());
			if (testproc(sid)==-1) {
#ifdef DEBUG
				char buffer[80];
				sprintf(buffer,"Commdslave %s",SLAVE_DEATH);
				fprintf(stderr,"%s",buffer);
#endif DEBUG
				kc->processor[slave]=0;
				Alloc_wait(wait);
				return(-1);
			}
			Alloc_wait(wait);
		}
		tsleep(10);
	}
#endif
	return(1);
}
