/*
 *	File:	conveither.c
 *		
 *		Microscope coordinate conversion executable module.
 *
 *
 *	Written by:	S. C. Georgiou
 *				720 Birchwood Boulevard,
 *				Birchwood Science Park,
 *				Warrington,
 *				Cheshire.
 *
 *	Date: 4 Dec 1986.
 *
 *
 *	Modifications:
 *	--------------
 *
 *		When:		Who:		What:
 *		-----		----		-----
 *
 *	 7 Nov 1988		SEAJ		Changed to allow for changes to mic_conv.c
 *	12 Oct 1988		SEAJ		Revamp to coinside with mic_conv.c
 *	16 Apr 1986		CAS			Change coord file name + can't open message
 *	15 Apr 1987		SCG			A10 ref. point instead of A5.
 *	14 Apr 1987		SCG			Read mic coords from file. mm instead of mu.
 *	17 Mar 1987		SCG			Tidied up the code. Introduced LH, RH  
 *								stage philosophy.  Made code more
 *								efficient.
 *	05 Mar 1987		SCG			Further mods to cope with 180 rotation
 *								of slide w.r.t. NE finder on target mic.
 *	03 Mar 1987		SCG			Put options for different types of
 *								mics and the ability to measure slide width.
 *	19 Feb 1987		SCG			Cludge option under devel mode.
 *	12 Feb 1987		SCG			Modified functionality of this to agree with
 *								mic_conv.c in /dd/usr/sotiris/src/trans_coord
 *	16 Dec 1986		SCG			Modification to prevent hanging when both
 *								sets of coordinates are zero.
 *
 *
 */
#define	INIFILE
#include	<cmdsys.h>
#include	<math.h>

/*
 *	Local Defines
 */
#define MSCOPEFILE "/dd/params/mscopecoords"
#define ERRFILE "micerrfile"
#define MAXNAME 16
#define DIRSTART 5
#define SCREENFUL 15
#define MAXOPTS 4		/* maximum number of options for conversion */
#define RADIUS 250		/* radius of circle on NE finder */
#define RADIUS2 62500	/* radius squared */
#define DIST 338		/* distance to center of quadrants */
#define CONVOFF 0		/* type of conversion */
#define MICROSCOPE 1
#define ENGLISHNE 2
#define SWEDISHNE 3

/* structure definitions */
static struct coord {
	int x;
	int y;
};

static struct microscope {
	int number;
	struct coord Z50;
	struct coord A10;
	int left_handed;
	int inverted;
	char id[MAXNAME];
} mic = {0, { 0, 0},{ 0, 0}, 1, 1, 'U', 'N', 'S', 'E', 'T'};

static char *stage[] = {
				"Right-handed",
				"Left-handed"
			};
static char *optics[] = {
				"Non-inverted",
				"Inverted"
			};
/* global defines */
static struct coord cytoZ50 = { -11648, -11674 },	/* Cytoscan Z50 coordinates */
					cytoA10 = { 11649, 27149},		/* Cytoscan A10 coordinates*/
					cytoM30 = { 902, 7727 },		/* Cytoscan M30 coordinates for check */
					middle;							/* coordinates of midpoint of slide
																on target mic */
																
static	float	sin1, cos1, 		/* First rotation angle sin & cos. */
				sin2, cos2; 		/* Second rotation angle sin & cos. */
static	int		xoff, yoff;			/* Translation offsets. */
                                       
static	int 	iflag;				/* Set when coords reflected. */
static	int		ccoord = 0;			/* Type of coordinat conversion */
static	int 	slwidth = 25000;	/* Slide width */
static	int		scanright = 12950;	/* From datum to right edge */

static	int	debug = 0;
static	FILE *debug_channel;
/* Forward declarations. */
static	int mic_conv(), select_mic(), micread(), shownames(),readandconvert(),
			setiflag(), eval_ang(), eval_mids(), backconv(),
			convertback(), conv_coords(), cyto_ne(), ne_cyto(), cyto_sne(),
			sne_cyto(), offset(), rotate(), nextline(), checkstr(), chekstr(),
			checkref(), checkfil(), float_str(), conv_str(), isnumber();
static	float atan2();
static  char *skipword(), *skipwhite();
/*
 *
 *  C O N V E I T H E R --
 *
 *			Obtain microscope coordinates of the datum point or optionally of 
 *			two points to use in the conversion from local to microscope
 *			coordinates. 
 */
main()
{
	int i, cmd, Quitcmd;
	int option_is[MAXOPTS];
	char micname[MAXNAME];
	FILE *fopen();
	
	muini(stdout);
	if ((debug_channel = fopen(ERRFILE, "w")) == NULL) 
		debug = 0;
		
	for (i = 0; i < MAXOPTS; i++)
		option_is[i] = 0;
	option_is[ccoord] = 1;
	
	newmenu("Coordinate Conversion");
	mmcmd = 0;
	cmmd("coordinate conversion off;:k40%-t", &option_is[mmcmd]);
	cmmd("microscope verniers;:k40%-t", &option_is[mmcmd]);
	cmmd("english england finder;:k40%-t", &option_is[mmcmd]);
	cmmd("swedish england finder;:k40%-t", &option_is[mmcmd]);
	cmmd("select microscope;%?+s%f", micname, select_mic);
	cmmd("single case backward conversion;%f", backconv);
	cmmd("convert from file;%f", readandconvert);
	cmmd("quit"); Quitcmd = mmcmd;
	
	while ((cmd = thismu()) != Quitcmd) {
		if (cmd > 0 && cmd <= 4) {
			ccoord = cmd - 1;
			for (i=0; i<MAXOPTS; i++)
				option_is[i] = 0;
			option_is[ccoord] = 1;
			for (i = 1; i <= 4; i++)
				displin(0, i);
			putprompt();
		}
	}
	alldone(0);	
}

/*
 *  S E L E C T _ M I C -- Obtain microscope coordinates of the datum point or 
 *						   optionally of two points to use in the conversion 
 *						   from local to microscope coordinates. 
 */
select_mic(string)
char *string;
{
int  Scmd, Icmd, Wcmd, i;
char Z50x[9], Z50y[9], A10x[9], A10y[9], lenstr[9], tmpstr[9], langstr[10];
char stagestr[13], opticsstr[13], slidestr[13];

	/* Convert slide width to string. */
	float_str(lenstr, slwidth);
	/* Convert coords to strings. */
	float_str(Z50x, mic.Z50.x);
	float_str(Z50y, mic.Z50.y);
	float_str(A10x, mic.A10.x);
	float_str(A10y, mic.A10.y);
	strcpy(stagestr, stage[mic.left_handed]);
	strcpy(opticsstr, optics[mic.inverted]);
	*tmpstr = '\0';
	if (!strcmp(string, "UNSET") || !micread(string)) {
		newmenu("Select Microscope");
		cmmd("microscope id;:k40 %+s%-s", string, mic.id);  Icmd = mmcmd;
		cmmd("show microscope directory;%f", shownames);
		cmmd("microscope coords for Z50;:k40 x %8s y %8s %-+d %-+d %f", Z50x, 
										Z50y, &mic.Z50.x, &mic.Z50.y, checkstr);
		cmmd("microscope coords for A10;:k40 x %8s y %8s %-+d %-+d %f", A10x, 
										A10y, &mic.A10.x, &mic.A10.y, checkstr);
		cmmd("stage;:k40%+T%-s%-+d%f", &mic.left_handed, stagestr, stage, conv_str);
		cmmd("optics;:k40%+T%-s%-+d%f", &mic.inverted, opticsstr, optics, conv_str);
		cmmd("slide width;:k40%+s%-s", tmpstr, lenstr); Wcmd = mmcmd;
	
		while(thismu()) {
			switch(nncmd) {
				case 1:
					if(micread(string)) {
						float_str(Z50x, mic.Z50.x);
						float_str(Z50y, mic.Z50.y);
						float_str(A10x, mic.A10.x);
						float_str(A10y, mic.A10.y);
						strcpy(stagestr, stage[mic.left_handed]);
						strcpy(opticsstr, optics[mic.inverted]);
						for (i = 0; i <= 6; i++)
							displin(0, i);
					} else {
						strcpy(string, "UNSET");
						displin(0, Icmd);
						screengo(ENDLINE, 0);
						fprintf(output, "Microscope id match not found.\007\n");
					}
					putprompt();
					break;
				case 7:
					if(!chekstr(tmpstr) || (*tmpstr == '-') || (*tmpstr == '\0')) {/* Check for stupid strings. */
						screengo(ENDLINE, 0);
						fprintf(output, "Illegal value.\n");
					} else {
						slwidth = (int) (1000.0 * atof(tmpstr));
						strcpy(lenstr, tmpstr);
						screengo(ENDLINE, 0);
						displin(0, Wcmd);
						putprompt();
					}
					break;
			}
		}
	}
	if (ccoord == MICROSCOPE) {
		if(mic.Z50.x == mic.A10.x && mic.Z50.y == mic.A10.y) {	/* Bad coords given.*/
			fprintf(output, "Coords define same point.  No conversion.\n");
			ccoord = 0;
			return;
		}
	} else if (ccoord > MICROSCOPE) {
		if (mic.Z50.x == 0) {
			mic.Z50.x = cytoZ50.x;
			mic.Z50.y = cytoZ50.y;
			mic.A10.x = cytoA10.x;
			mic.A10.y = cytoA10.y;
			if (!mic.inverted) {			/* non inverted mic */
				mic.inverted = 1;			/* make it inverted */
				if (mic.left_handed)		/* left handed mic */
					mic.left_handed = 0;	/* make it right handed */
				else						/* right handed mic */
					mic.left_handed = 1;	/* make it left handed */
			}
		}
	}
}

/*
 *	M I C R E A D  -- Read microscope coords from file given microscope id.
 *
 */
static
int	micread(name)
char *name;
{
	int fd, i;
	char buffer[200], *buf, str[MAXNAME];

	if ((fd = open(MSCOPEFILE, 1)) == -1 ) {
		fprintf(output, "Unable to read microscope file\n");
		return(0);
	}

	while(nextline(fd, buffer, sizeof(buffer))) {
		buf = buffer;
		buf = skipwhite(buf);
		if (isdigit(*buf)) {
			mic.number = atoi(buf);
			buf = skipword(buf);
			i = 0;
			while ((*buf != ' ') && (*buf != '\t')) {
				str[i] = *buf;
				buf++;
				i++;
			}
			str[i] = '\0';
			buf = skipwhite(buf);
			if (strcmp(name, str) == 0) {
				strcpy(mic.id, name);
				mic.Z50.x = (int) (atof(buf) * 1000 );
				buf = skipword(buf);
				mic.Z50.y = (int) (atof(buf) * 1000 );
				buf = skipword(buf);
				mic.A10.x = (int) (atof(buf) * 1000 );
				buf = skipword(buf);
				mic.A10.y = (int) (atof(buf) * 1000 );
				buf = skipword(buf);
				if (strncmp(buf, "LH", 2) == 0)
					mic.left_handed = 1;
				else
					mic.left_handed = 0;
				buf = skipword(buf);
				if (strncmp(buf, "INV", 3) == 0)
					mic.inverted = 1;
				else
					mic.inverted = 0;
				strcpy(name, "UNSET");
				close(fd);
				return(1);
			} /* end of if */
		} /* end of if isdigit */
	} /* end of while */
	strcpy(name, "UNSET");
	close(fd);
	return(0);  /* no match */
}
/*
 * S H O W N A M E S -- Show the names of all the microscopes known
 *
 */
shownames()
{
	int fd, number, count, flag=0;
	char buffer[200], *buf, *str;
	
	if ((fd = open(MSCOPEFILE, 1)) == -1) {
		fprintf(output, "Unable to open microscopes file\n");
		return(0);
	}
	vclear(ENDLINE);
	screengo(1, 20);
	printf("Directory of Microscopes\n\n");
	screengo(3, 0);
	printf("\t\tNumber\t\t\tName\n\n");
	scrollpart(DIRSTART, ENDLINE);
	screengo(DIRSTART, 0);
	count = 0;
	while(nextline(fd, buffer, sizeof(buffer))) {
		buf = buffer;
		buf = skipwhite(buf);
		if (isdigit(*buf)) {
			count++;
			number = atoi(buf);
			buf = skipword(buf);
			str = skipword(buf);
			str--;
			*str = '\0';
			printf("\t\t%d\t\t\t%s\n", number, buf);
			if (count >= SCREENFUL) {
				count = 0;
				if (!go_on()) {
					flag = 1;
					break;
				}
				clearlns(DIRSTART, ENDLINE);
				screengo(DIRSTART, 0);
			}
		}
	}
	if (!flag)
		if (go_on());
	vclear(ENDLINE);
	scrollpart(OUTLINE, ENDLINE);
	setdisplay(BUST);
}
/*
 *  R E A D A N D C O N V E R T  --  Microscope to Cytoscan from file.
 *
 */
static
readandconvert()
{
struct coord pt, cyto;
int cmd, Excmd, fd, ok;
static char	infilnam[16], outfilnam[16];
char string[6], buffer[200], *buf;
FILE	*fopen(), *fpout;

	newmenu("Backward Conversion from file");
	cmmd("Input file;:k40 %s%f", infilnam, checkfil);
	cmmd("Output file;:k40 %s%f", outfilnam, checkfil);
	cmmd("Exit"); Excmd = mmcmd;
		
	while((cmd = thismu()) != Excmd) {
		if(!nncmd) {
			if (ccoord == MICROSCOPE) {
				if(mic.Z50.x == mic.A10.x && mic.Z50.y == mic.A10.y) {
					fprintf(output, "Z50 A10 coords define same point.  No conversion.\n");
					putprompt();
						ok = 0;
				} else if (!strcmp(mic.id, "UNSET")) {
					fprintf(output, "Microscope is UNSET. No conversion.\n");
					ok = 0;
				} else
					ok = 1;
				} else if ((ccoord > MICROSCOPE) && (!strcmp(mic.id, "UNSET"))) {
				mic.Z50.x = cytoZ50.x;
				mic.Z50.y = cytoZ50.y;
				mic.A10.x = cytoA10.x;
				mic.A10.y = cytoA10.y;
				/* make microscope one of the two defaults */
				if (!mic.inverted) {			/* non inverted mic */
					mic.inverted = 1;			/* make it inverted */
					if (mic.left_handed)		/* left handed mic */
						mic.left_handed = 0;	/* make it right handed */
					else						/* right handed mic */
						mic.left_handed = 1;	/* make it left handed */
				}
				ok = 1;
			}
			if (ok) {
				setiflag();
				eval_ang();
				eval_mids();
				if((fd = open(infilnam, 1)) != -1) {
					if((fpout = fopen(outfilnam, "w")) != NULL) {
						fprintf(fpout,"*\n* Cytoscan coordinates in microns\n*\n");
						fprintf(fpout,"*      x         y\n");
						while(nextline(fd, buffer, sizeof(buffer))) {
							buf = buffer;
							buf = skipwhite(buf);
							if (ccoord == CONVOFF) {
								fprintf(output, "Type of conversion not set.\n");
								putprompt();
								break;
							}
							switch(ccoord) {
								case CONVOFF :
									break;
								case MICROSCOPE :
									pt.x = (int) (1000.0 * atof(buf));
									pt.y = (int) (1000.0 * atof(buf));
									convertback(pt.x, pt.y, &cyto.x, &cyto.y);
									break;
								case ENGLISHNE :
									strcpy(string, buf);
									if (checkref(string))
										ne_cyto(string, &cyto.x, &cyto.y);
										break;
								case SWEDISHNE :
									strcpy(string, buf);
									if (checkref(string))
										sne_cyto(string, &cyto.x, &cyto.y);
									break;
							}
							if (ccoord != CONVOFF)
								fprintf(fpout,"  %8d  %8d\n", cyto.x, cyto.y);
						}
						close(fd);
						fclose(fpout);
					} else
						fprintf(output,"Unable to open output file %s\n",outfilnam);				
				} else
				fprintf(output,"Unable to open input file %s\n",infilnam);				
			}
			putprompt();
		}
	}
}
/*
 *  S E T I F L A G  --
 * 
 *				Sets iflag.  Determines whether microscope reflecting or not. 
 */
static
setiflag()
{
int	d1, d2;

	d1 = mic.Z50.y - mic.A10.y;
	d2 = mic.Z50.x - mic.A10.x;
	if(mic.inverted) {			/* Inverted microscope. */
	
		if((d1 < 0 && d2 > 0) || (d1 > 0 && d2 < 0)) {	/* Reflecting. */
			iflag = 1;
		} else 											/* Non reflecting. */
			iflag = 0;
			
	} else {				/* Non inverted microscope. */

		if((d1 < 0 && d2 < 0) || (d1 > 0 && d2 > 0)) {	/* Reflecting. */
			iflag = 1;
		} else 											/* Non reflecting. */
			iflag = 0;
			
	}
}
/*
 *  E V A L _ A N G  --
 * 
 *				Calculates angles and offsets required for transformation. 
 *				Actual rotation angle theta2 - theta1.  Used two angles
 *				to make sure that sign is right.
 */
static
eval_ang()
{
float	gly, glx, lcx, lcy, theta, theta1, theta2, Pi;
int		xpos, ypos;

	if(iflag) {							/* Consider reflection. */
		mic.Z50.x = -mic.Z50.x;
		mic.A10.x = -mic.A10.x;
	}
	
	gly = (float) (mic.Z50.y - mic.A10.y);	/* Evaluate first angle. */
	glx = (float) (mic.Z50.x - mic.A10.x);
	theta1 = atan2(gly, glx);
	sin1 = (float) sin(theta1);
	cos1 = (float) cos(theta1);
	
	lcy = (float) (cytoZ50.y - cytoA10.y);	/* Evaluate second angle. */
	lcx = (float) (cytoZ50.x - cytoA10.x);
	theta2 = atan2(lcy, lcx);
	sin2 = (float) sin(theta2);
	cos2 = (float) cos(theta2);

	/* Evaluate offsets. */
	offset((float) mic.Z50.x,(float) mic.Z50.y,(float) cytoZ50.x,(float) cytoZ50.y);
		   
	if(iflag) {							/* Correct for reflection. */
		mic.Z50.x = -mic.Z50.x;
		mic.A10.x = -mic.A10.x;
	}
	
	if(debug) {
		Pi = 4 * atan(1.);
		theta = (theta2 - theta1) * 180 / Pi;
		conv_coords(cytoM30.x, cytoM30.y, &xpos, &ypos);	
		fprintf(debug_channel, "Rotation angle %f\n", theta);
		fprintf(debug_channel, "Translation in x %d and y %d\n", xoff, yoff);
		fprintf(debug_channel, "Transformed point at M30 %d %d\n", xpos, ypos);
	}
		   
}
/*
 *  E V A L _ M I D S  --
 * 
 *				Calculates coords of a point of slide on target mic about 
 *				which the slide is rotated.  When the slide is horiz. this
 *				point's x coord is half way along the length of the NE
 *				finder and its y coord half way along the width of the slide.
 *				If slide vertical coords are other way round.  The slide is
 *				rotated when transforming to a LH stage (inverted mic)
 *				or a RH stage (non-inverted mic).
 */
static
eval_mids()
{
float	n38x, n38y, d1, d2;

	d1 = (float) (mic.Z50.x - mic.A10.x);
	d2 = (float) (mic.Z50.y - mic.A10.y);
	
	if(fabs(d1) > fabs(d2)) {				/* Slide horizontal. */
											/* 28 - Diff between A10 N38 */
											/* 40 - Diff between A10 Z50 */
											/* 12 - Diff between A10 N38 */
											/* 24 - Diff between A10 Z50 */
		n38x = d1 * 28 / 40 + (float) mic.A10.x;
		n38y = d2 * 12 / 24 + (float) mic.A10.y;
		middle.x = (int) n38x;
		if(d2 > 0) 
			middle.y = (int) n38y - (scanright - slwidth / 2);
		else
			middle.y = (int) n38y + (scanright - slwidth / 2);
			
	} else {							/* Slide vertical. */
	
		n38x = d1 * 12 / 24 + (float) mic.A10.x;
		n38y = d2 * 28 / 40 + (float) mic.A10.y;
		if(d1 > 0)
			middle.x = (int) n38x - (scanright - slwidth / 2);
		else
			middle.x = (int) n38x + (scanright - slwidth / 2);
		middle.y = (int) n38y;
	}
	if(debug) {
		fprintf(debug_channel, "middle.x %d middle.y %d\n", middle.x, middle.y);
	}
}

/*
 *  B A C K C O N V  --  Microscope to Cytoscan conversion.
 *
 */
static
backconv()
{
static struct coord mscope, cytoscan;
char	strx[8], stry[8], NEstr[8], SNEstr[8];
int ok, Cytocmd, Backcmd;


	/* Convert coords to strings. */
	float_str(strx, mscope.x);
	float_str(stry, mscope.y);
	NEstr[0] = '\0';
	SNEstr[0] = '\0';
	
	screengo(ENDLINE, 0);
	fprintf(output, "Enter Microscope/England Finder Reference, and <return>.\n");
	newmenu("Backward Conversion");
	switch (ccoord) {
		case MICROSCOPE :
			cmmd("microscope coords;:k40 x %8s y %8s%-+d%-+d%f", strx, stry, 
								&mscope.x, &mscope.y, checkstr); Backcmd = mmcmd;
			break;
		case ENGLISHNE :
			cmmd("england finder (Graticules Ltd.);:k40 %s", NEstr); Backcmd = mmcmd;
			break;
		case SWEDISHNE :
			cmmd("england finder (Albert Sass);:k40 %s", SNEstr); Backcmd = mmcmd;
			break;
	} 	/* end of switch */
	cmmd("cytoscan coords;:k40 x %d y %d", &cytoscan.x, &cytoscan.y); Cytocmd = mmcmd;
	while(thismu()) {
		if (ccoord == MICROSCOPE) {
			if(mic.Z50.x == mic.A10.x && mic.Z50.y == mic.A10.y) {
				fprintf(output, "Z50 A10 coords define same point.  No conversion.\n");
				putprompt();
				ok = 0;
			} else if (!strcmp(mic.id, "UNSET")) {
				fprintf(output, "Microscope is UNSET. No conversion.\n");
				ok = 0;
			} else
				ok = 1;
		} else if ((ccoord > MICROSCOPE) && (!strcmp(mic.id, "UNSET"))) {
			mic.Z50.x = cytoZ50.x;
			mic.Z50.y = cytoZ50.y;
			mic.A10.x = cytoA10.x;
			mic.A10.y = cytoA10.y;
			/* make microscope one of the two defaults */
			if (!mic.inverted) {			/* non inverted mic */
				mic.inverted = 1;			/* make it inverted */
				if (mic.left_handed)		/* left handed mic */
					mic.left_handed = 0;	/* make it right handed */
				else						/* right handed mic */
					mic.left_handed = 1;	/* make it left handed */
			}
			ok = 1;
		}
		if (ok) {
			setiflag();
			eval_ang();
			eval_mids();
			if (nncmd == Cytocmd) {
				switch (ccoord) {
					case MICROSCOPE :
						conv_coords((short) cytoscan.x, (short) cytoscan.y, &mscope.x, &mscope.y);
						float_str(strx, mscope.x);
						float_str(stry, mscope.y);
						break;
					case ENGLISHNE :
						if (checkref(NEstr))
							cyto_ne((short) cytoscan.x, (short) cytoscan.y, NEstr);
						break;
					case SWEDISHNE :
					if (checkref(SNEstr))
							cyto_sne((short) cytoscan.x, (short) cytoscan.y, SNEstr);
						break;
					default :
						ok = 0;
						break;
				} 	/* end of switch */
				if (ok)
					displin(0,Backcmd);
			} else if (nncmd == Backcmd) {
				switch (ccoord) {
					case MICROSCOPE :
						convertback(mscope.x, mscope.y, &cytoscan.x, &cytoscan.y);
						break;
					case ENGLISHNE :
						if (checkref(NEstr))
							ne_cyto(NEstr, &cytoscan.x, &cytoscan.y);
						break;
					case SWEDISHNE :
						if (checkref(SNEstr))
							sne_cyto(SNEstr, &cytoscan.x, &cytoscan.y);
						break;
					default :
						ok = 0;
						break;
				} 	/* end of switch */
				if (ok)
					displin(0,Cytocmd);
			}
			putprompt();
		} /* end of if ok */
	} /* end of while */
}
/*
 *
 *  C O N V E R T 	B A C K  -- Microscope to Cytoscan conversion.
 *
 */ 
static 
convertback(oldx, oldy, newx, newy)
int	oldx, oldy; 
int	*newx, *newy;
{
float  nx, ny, rot1x, rot1y, rot2x, rot2y, bix, biy;

	if(debug) {
		fprintf(debug_channel, "In converback.\n");
	}
	if((mic.inverted && mic.left_handed) || (!mic.inverted && !mic.left_handed)) {	/* Correct for 180 deg 
													relative rotation. */
		nx = (float) (oldx - middle.x);				/* Subtract offsets. */
		ny = (float) (oldy - middle.y);
													/* Rotate by Pi rads. */
		rotate(&rot1x, &rot1y, nx, ny, 0.0, -1.0);	
		rot1x = rot1x + (float) middle.x;			/* Add offsets. */
		rot1y = rot1y + (float) middle.y;
		
	} else {										/* No relative rotation. */
		rot1x = (float) oldx;
		rot1y = (float) oldy;
	}
	if(iflag)
		rot1x = -rot1x;
	rot1x -= (float) xoff;
	rot1y -= (float) yoff;
	rotate(&rot2x, &rot2y, rot1x, rot1y, sin2, cos2);
	rotate(&bix, &biy, rot2x, rot2y, -sin1, cos1);
	*newx = (int) bix;
	*newy = (int) biy;
	if(debug) {
		fprintf(debug_channel, "Out of conv_coords.\n");
	}
}
/*
 *  C O N V _ C O O R D S  -- Converts from cytoscan coordinates in microns to 
 *							  microscope coordinates in microns.
 */ 
static 
conv_coords(oldx, oldy, newx, newy)
short	oldx, oldy; 
int	*newx, *newy;
{
float  nx, ny, rot1x, rot1y, rot2x, rot2y;

	if(debug) {
		fprintf(debug_channel, "In conv_coords.\n");
	}

	rotate(&nx, &ny, (float) oldx, (float) oldy, sin1, cos1);
	
	if(debug) 
		fprintf(debug_channel,"First trans: %f %f\n", nx, ny);
		
	rotate(&rot1x, &rot1y, nx, ny, -sin2, cos2);
	
	if(debug) 
		fprintf(debug_channel,"Second trans: %f %f\n", rot1x, rot1y);
		
	rot1x += (float) xoff;
	if(iflag) {
		rot1x = -rot1x;
	}
	rot1y += (float) yoff;
	
	if(debug) 
		fprintf(debug_channel,"Third trans: %f %f\n", rot1x, rot1y);
		
	if((mic.inverted && mic.left_handed) || (!mic.inverted && !mic.left_handed)) {	/* Correct for 180 deg 
													relative rotation. */
		rot1x = rot1x - (float) middle.x;				/* Subtract offsets. */
		rot1y = rot1y - (float) middle.y;
		rotate(&rot2x, &rot2y, rot1x, rot1y, 0.0, -1.0);	/* Rotate by Pi rads. */
		rot2x = rot2x + (float) middle.x;			/* Add offsets. */
		rot2y = rot2y + (float) middle.y;
		*newx = (int) rot2x;
		*newy = (int) rot2y;
		
	} else {										/* No relative rotation. */
	
		*newx = (int) rot1x;
		*newy = (int) rot1y;
	}
	
	if(debug) 
		fprintf(debug_channel,"Fourth trans: %d %d\n", *newx, *newy);
	
	if(debug)
		fprintf(debug_channel, "Out of conv_coords.\n");
}
/*
 * C Y T O _ N E -- Cytoscan coordinates to New England
 *
 */
cyto_ne(x, y, string)
short x, y;				/* cytoscan coords */
char *string;
{
	static struct coord scope, new, n, sub;
	int d1, d2, xd, yd, quad, tmp;
	char ch;
	
	/* convert cytoscan coords to microscope coords */
	conv_coords(x, y, &scope.x, &scope.y);
	if (debug)
		fprintf(debug_channel, "x = %d y = %d scope = (%d, %d)\n", x, y, scope.x, scope.y);
	d1 = mic.Z50.x - mic.A10.x;
	d2 = mic.Z50.y - mic.A10.y;
	
	xd = scope.x - mic.A10.x;
	yd = scope.y - mic.A10.y;
	if (debug)
		fprintf(debug_channel, "d1 = %d d2 = %d xd = %d yd = %d\n", d1, d2, xd, yd);
	
	if (abs(d1) > abs(d2)) {						/* slide horizontal */
		new.x = ((xd * 40000) / d1) + 10000;
		new.y = ((yd * 24000) / d2);
	} else {										/* slide vertical */
		new.x = ((xd * 24000) / d1);
		new.y = ((yd * 40000) / d2) + 10000;
	}		
	
	n.x = new.x / 1000;	/* letter */
	n.y = new.y / 1000;	/* number */
	
	/* calculate fractional remainder */
	sub.x = new.x - (n.x * 1000);
	sub.y = new.y - (n.y * 1000);
	if (abs(d1) > abs(d2)) {					/* slide horizontal */
		tmp = sub.x;
		sub.x = sub.y;
		sub.y = tmp;
		tmp = n.x;								/* swap n(x, y) so that number */
		n.x = n.y;								/* and letter are correct */
		n.y = tmp;
	}
	if (square(sub.x, sub.y) <= RADIUS2)
		quad = 0;
	else {
		if (sub.x <= 500) {
			if (sub.y <= 500)
				quad = 4;
			else {
				n.y++;				/*square above */
				sub.y = 1000 - sub.y;
				if (square(sub.x, sub.y) <= RADIUS2) quad = 0;
				else
					quad = 3;
			}
		} else {					/* x > 500 */
			if (sub.y < 500) {
				n.x++;				/* square above */
				sub.x = 1000 - sub.x;
				if (square(sub.x, sub.y) < RADIUS2) quad = 0;
				else
					quad = 2;
			} else {				/* square diagonally above */
				n.x++;
				n.y++;
				sub.x = 1000 - sub.x;
				sub.y = 1000 - sub.y;
				if (square(sub.x, sub.y) <= RADIUS2) quad = 0;
				else
					quad = 1;
			}
		}
	}
	if (debug)
	 	fprintf(debug_channel, "new  = (%d, %d) n = (%d, %d) sub = (%d, %d) ", 
					new.x, new.y, n.x, n.y, sub.x, sub.y);
	if (n.x < 8)
		ch = n.x + 'A';
	else
		ch = n.x + 1 + 'A';
	sprintf(string, "%c%d/%d", ch, n.y, quad);
}
/*
 *  N E _ C Y T O -- New England references to Cytoscan coords
 *
 */
ne_cyto(string, cytox, cytoy)
char *string;
int *cytox, *cytoy;
{
	static struct coord new;
	int x, y, quad, d1, d2, xd, yd;
	char *str;

	if (debug)
		fprintf(debug_channel, "string = %s cytox = %d cytoy = %d\n", string, *cytox, *cytoy);
	/* convert string into x, y, quad */
	while(*string == ' ')
		string++;			/* delete white space */
	if (*string <= 'I')
		x = *string - 'A';
	else
		x = *string - 'A' - 1;
	string++;
	str = string;
	while (*string != '/' && *string != '\0')
		string++;
	*string = '\0';
	y = atoi(str);
	string++;
	quad = atoi(string);
	if (debug)
		fprintf(debug_channel, "x = %d y = %d quad = %d\n", x, y, quad);
	
	d1 = mic.Z50.x - mic.A10.x;
	d2 = mic.Z50.y - mic.A10.y;
	
	if (abs(d1) > abs(d2)) {				/* slide horizontal */
		xd = y - 10;
		yd = x;
		new.x = (xd * d1) / 40 + mic.A10.x;
		new.y = (yd * d2) / 24 + mic.A10.y;
	} else {								/* slide vertical */
		xd = x;
		yd = y - 10;
		new.x = (xd * d1) / 24 + mic.A10.x;
		new.y = (yd * d2) / 40 + mic.A10.y;
	}
	if (debug)
	 	fprintf(debug_channel, "d1 = %d d2 = %d xd = %d yd = %d new = (%d, %d)\n", d1, d2, xd, yd, new.x, new.y);
	switch (quad) {
		case 0 :
			break;
		case 1 :
			new.y += DIST;
			if (abs(d1) > abs(d2))
				new.x -= DIST;
			else				
				new.x += DIST;
			break;
		case 2 :
			if (abs(d1) > abs(d2))
				new.y += DIST;
			else
				new.y -= DIST;
			new.x += DIST;
			break;
		case 3 :
			if (abs(d1) > abs(d2))
				new.y -= DIST;
			else				
				new.y += DIST;
			new.x -= DIST;
			break;
		case 4 :
			new.y -= DIST;
			if (abs(d1) > abs(d2))
				new.x += DIST;
			else
				new.x -= DIST;
			break;
	}
	convertback(new.x, new.y, cytox, cytoy);
	if (debug)
	 	fprintf(debug_channel, "new = (%d, %d) cyto = (%d, %d)\n", new.x, new.y, *cytox, *cytoy);
}
/*
 * C Y T O _ S N E -- Cytoscan coordinates to Swedish New England
 *					  NB  The whole of this routine is calculated in
 *						  10's of microns instead of microns to avoid
 *						  overflow problems.
 */
cyto_sne(x, y, string)
short x, y;
char *string;
{
	static struct coord scope, new, n;
	int d1, d2, xd, yd, tmp;
	char ch;
	
	/* convert cytoscan coords to microscope coords */
	conv_coords(x, y, &scope.x, &scope.y);
	if (debug)
	 	fprintf(debug_channel, "x=%d y=%d scope=(%d, %d)\n", x, y, scope.x, scope.y);
	
	d1 = (mic.Z50.x - mic.A10.x) / 10;
	d2 = (mic.Z50.y - mic.A10.y) / 10;
	
	xd = (scope.x - mic.A10.x) / 10;
	yd = (scope.y - mic.A10.y) / 10;
	
	if (abs(d1) > abs(d2)) {				/* slide horizontal */
		new.x = ((xd * 8090) / d1) + 2160;
		new.y = ((yd * 4850) / d2) + 150;
	} else {								/* slide vertical */
		new.x = ((xd * 4850) / d1) + 150;
		new.y = ((yd * 8090) / d2) + 2160;
	}	
	n.x = new.x / 100;	/* letter */
	n.y = new.y / 100;	/* number */
	if (abs(d1) > abs(d2)) {		/* swap number & letter if horizontal */
		tmp = n.x;
		n.x = n.y;
		n.y = tmp;
	}
	if (debug)
		fprintf(debug_channel, "d1=%d d2=%d xd=%d yd=%d new=(%d, %d) n=(%d, %d)\n", d1, d2, xd, yd, new.x, new.y, n.x, n.y);
	
	if (n.x <= 25)
		ch = n.x + 'A';
	else
		ch = n.x - 26 + 'a';
	sprintf(string, "%c%d", ch, n.y);
}
/*
 * S N E _ C Y T O -- Swedish New England references to Cytoscan coords
 *					  NB  All calculations are in 10's microns instead
 *						  of microns to stop overflow problems.
 */
sne_cyto(string, cytox, cytoy)
char *string;
int *cytox, *cytoy;
{
	static struct coord new;
	int x, y, d1, d2, xd, yd;
	
	while (*string == ' ')
		string++;
	if (*string <= 'Z')
		x = *string - 'A';
	else
		x = *string - 'a' + 26;
	string++;
	y = atoi(string);
	
	d1 = (mic.Z50.x - mic.A10.x) / 10;
	d2 = (mic.Z50.y - mic.A10.y) / 10;
	
	if (abs(d1) > abs(d2)) {				/* slide horizontal */
		yd = (x - 1) * 100;					/* x + 0.5 - 1.5 */
		xd = (y * 100) - 2110; 				/* y + 0.5 - 21.6 */
		new.x = (((xd * d1) / 8090) * 10) + mic.A10.x;
		new.y = (((yd * d2) / 4850) * 10) + mic.A10.y;
	} else {								/* slide vertical */
		xd = (x - 1) * 100;					/* x + 0.5 - 1.5 */
		yd = (y * 100) - 2110; 				/* y + 0.5 - 21.6 */
		new.x = (((xd * d1) / 4850) * 10) + mic.A10.x;
		new.y = (((yd * d2) / 8090) * 10) + mic.A10.y;
	}
	if (debug)
		fprintf(debug_channel, "x=%d y=%d d1=%d d2=%d xd=%d yd=%d new=(%d, %d)\n",
					x, y, d1, d2, xd, yd, new.x, new.y);
	convertback(new.x, new.y, cytox, cytoy);
}
/*
 * O F F S E T -- Rotate local to global and evaluate offsets.
 *
 */
static
offset(gx1, gy1, lx1, ly1)
float  gx1, gy1, lx1, ly1;
{
float	nx, ny, rot1x, rot1y;

		rotate(&nx, &ny, lx1, ly1, sin1, cos1);
		rotate(&rot1x, &rot1y, nx, ny, -sin2, cos2);
		xoff = (int) (gx1 - rot1x);
		yoff = (int) (gy1 - rot1y);
		
}
/*
 *	R O T A T E -- Rotate a point through angle theta.
 *
 */	
static
rotate(newx, newy, oldx, oldy, sinth, costh)
float	*newx, *newy, oldx, oldy, sinth, costh; 
{

	*newx = oldx * costh - oldy * sinth;
	*newy = oldx * sinth + oldy * costh;
	
}
/*
 *	A T A N 2 -- Proper arctan evaluation through 360 degrees/2*pi radians
 *
 */
float	atan2(Dy, Dx)
float	Dy, Dx;
{
float	ttht, Pi;
	
	Pi = 4 * atan(1.);
	
	if(!Dx) {
		if(Dy < 0) {
			return(-Pi/2);
		} else {
			return(Pi/2);
		}
	}
	ttht = atan(fabs(Dy/Dx));
	if(Dy/Dx <= 0) {
		if(Dx < 0) {
			return(Pi - ttht);
		} else {
			return(-ttht);
		}
	} else {
		if(Dy < 0) {
			return(Pi + ttht);
		}
	}
	return(ttht);
	
}
		
/*
 *	N E X T L I N E  --
 *
 */
static int nextline(fd,buffer,len)
int		fd, len;
char	*buffer;
{
	int		cnt;

	while ((cnt = readln(fd,buffer,len)) == 1)
		;
	if (cnt <= 0)
		return(0);
	else {
		buffer[cnt-1] = '\0';
		return(cnt);
	}
}

/*
 *	S K I P W O R D  --
 *
 */
static char *skipword(str)
char	*str;
{
	while (*str && (*str != ' ') && (*str != '\t'))
		str++;
	str = skipwhite(str);
	return(str);
}	

/*
 *	S K I P W H I T E  --
 *
 */
static char *skipwhite(str)
char	*str;
{
	while (*str && ((*str == ' ') || (*str == '\t')))
		str++;
	return(str);
}

/*
 *	C H E K S T R -- Check that sensible string is given.
 *
 */
static
int chekstr(mystr)
char	*mystr;
{
char	*cptr;

	cptr = mystr;
	if((*cptr) == '+') {			/* Remove a leading + if it's there. */
		do {
			*cptr = *(cptr+1);
		} while((*cptr++) != '\0');
		cptr = mystr;
	}
	if((*cptr) == '-') {			/* Skip over a leading minus. */
		cptr = mystr+1;
	}
		
	while((*cptr) != '\0') {		/* Check the rest for sensible values. */
		if(! isdigit(*cptr) && (*cptr) != '.') {
			sprintf(mystr, "0.000");
			return(0);
		}
		cptr++;
	}
	return(1);
}

/*
 * C H E C K S T R -- checks strings entered for faults 
 *
 */
checkstr(str1, str2, x, y)
char *str1, *str2;
int *x, *y;
{
	if(!chekstr(str1)) {		/* Check for stupid strings. */
		screengo(ENDLINE, 0);
		fprintf(output, "Illegal value.  Reset to 0\n");
	}
	if(!chekstr(str2)) {		/* Check for stupid strings. */
		screengo(ENDLINE, 0);
		fprintf(output, "Illegal value.  Reset to 0.\n");
	}
	*x = (int) (1000.0 * atof(str1));
	*y = (int) (1000.0 * atof(str2));
	sprintf(mic.id, "UNSET");	/* Mic id now irrelevant. */
}
/*
 * C H E C K R E F -- check input NE reference 
 *
 */
checkref(string)
char *string;
{
	char str[20], *str1, *str2;
	int num;
	
	strcpy(str, string);
	str1 = str;
	while (*str1 == ' ')
		str1++;
	if (ccoord == 2) {
		if (*str1 < 'A' || *str1 > 'Z' || *str1 == 'I') {
			fprintf(output, "Illegal first character\n");
			return(0);
		} else {
			str1++;
			str2 = str1;
			while(isdigit(*str1) && *str1 != '\0')
				str1++;
			if (*str1 == '\0') {
				if (isnumber(str2))
					num = atoi(str2);
				else
					num = -1;
			} else {
				*str1 = '\0';
				if (isnumber(str2))
					num = atoi(str2);
				else
					num = -1;
				str1++;
			}
			if (num < 0 || num > 50) {
				fprintf(output, "Illegal number\n");
				return(0);
			} else {
				if (*str1 == '\0')
					return(1);
				if (isnumber(str1))
					num = atoi(str1);
				else
					num = -1;
				if (num < 0 || num > 4) {
					fprintf(output, "Illegal quadrant number\n");
					return(0);
				}
			}
		}
		return(1);
	} else {
		if ((*str1 >= 'a' && *str1 <= 'z') || (*str1 >='A' && *str1 <= 'Z')) {
			str1++;
			if (isnumber(str1))
				num = atoi(str1);
			else
				num = -1;
			if (num < 20 || num > 120) {
				fprintf(output, "Illegal number\n");
				return(0);
			} else
				return(1);
		} else {
			fprintf(output, "Illegal character\n");
			return(0);
		}
	}
}
/*
 * C H E C K F I L -- check filename entered
 *
 */
checkfil(str)
char *str;
{
	if (!isalpha(*str)) {
		screengo(ENDLINE, 1);
		fprintf(output, "Illegal name. Reset to NULL.\n");
		str = '\0';
	}
} 
/*
 * F L O A T _ S T R -- convert float number to string
 *
 */
float_str(string, number)
char *string;
int number;
{
	int num1, num2, negative = 0, five = 0;
	
	if (number < 0) {
		negative = 1;
		number *= -1;
	}
	num1 = number / 1000;
	num2 = number - (num1 * 1000);		/* decimal part of number */
	num2 /= 10;
	if ((num2 % 5) != 0) {
		num2 += 5;
		num2 /= 10;
		if (num2 >= 10) {
			num1++;
			num2 = 0;
		}
	} else {
		if ((num2 % 10) == 0)
			num2 /= 10;						/* remove trailing zero */
		else
			five = 1;						/* keep trailing 5 */
	}
	
	if (negative) {
		if (five && num2 < 10)
			sprintf(string, "-%d.0%d", num1, num2);
		else
			sprintf(string, "-%d.%d", num1, num2);
	} else {
		if (five && num2 < 10)
			sprintf(string, "%d.0%d", num1, num2);
		else
			sprintf(string, "%d.%d", num1, num2);
	}
}
/*
 * C O N V _ S T R
 *
 */
conv_str(num, str, array)
int *num;
char *str, *array[];
{
	strcpy(str, array[*num]);
}		

/*
 * I S N U M B E R --
 *
 */
 static int isnumber(string)
 char *string;
 {
 	char *str;
 
	str = string;	
 	while(*str != '\0') {
 		if (!isdigit(*str))
 			return(0);
 		else
 			return(1);
 	}
 }
/*
 * S Q U A R E -- square two integer and return the sum
 *
 */
int square(x, y)
int x, y;
{
	int squ;
	
	squ = (x * x) + (y * y);
	return(squ);
}
/*
 *	S T R C A S E C M P  --  Compare two strings but ignore case
 *							 differences and top bit differences
 */
strcasecmp(s1, s2)
char *s1, *s2;
{
	while (*s1 && *s2) {
		if (toupper(*s1 & 0x7f) != toupper(*s2 & 0x7f))
			break;
		s1++; s2++;
	}
	return(toupper(*s1 & 0x7f) - toupper(*s2 & 0x7f));
}
gethelp() {}		
