/*
 * split_overs.c
 * -------------
 * Ji Liang June 18, 1987
 *
 * split overlap chromosomes
 * return 0 : failed
 *	  1 : successed
 * =====================
 * Modification history
 * --------------------
 *	29 Nov 1988		SCG			return/fail when overlap of joined objects requested
 *	22 Nov 1988		dcb			Added param. to woolz_check.. to indicate where called from
 *	16 Nov 1988		SCG			woolz_check_obj on entry
 *	24 Oct 1987		CAS			Debugging
 *	25-09-87		jimp@IRS	Get yesterday's bug-fix to work !!
 *	24-09-87		jimp@IRS	Check that the four split-points are distinct
 *								(function check_four).
 *	09-09-87		jimp		Minor mods around splithole.  Lower size limit.
 *	04-09-87		jimp		Check number of holes before proceding.  Call
 *								mendhole for small holes, as in split_comps.
 *	26-08-87		jimp		Mods to permit space freeing after splithole()
 *	18-08-87		jimp		Boundaries not displayed here
 *	16-08-87		jl			Changed disoverlap() from object to int.
 *	05-08-87		jimp		"point" again made local to avoid name clashes (see
 *								comment 29-07-87)
 *	04-08-87		JL			Use ibound() instead of boundary(). Unified the
 *								the functoin with woolz.
 *	01-07-87		jl		 	If there is too noisy, rethreshold.
 *								Holes may be mended or split at first.
 *	29-07-87		jimp		"point" made local to avoid name clashes.
 */
#include <stdio.h>
#include <wstruct.h>
#include <split.h>

#ifdef DEBUG
static FILE *debug_channel;
#endif

/*
 *	S P L I T _ O V E R S  --
 *
 */
split_overs(comobj, new0, new1, fr, col0, col1)
struct object *comobj, **new0, **new1;
struct pframe *fr;
int col0, col1;
{
	struct object *obj, *sobj, *mobj;
	struct object *splithole(), *mendhole();
	struct object *bobj, *ibound(), *subobj[2];
	struct boundlist *blist;
	struct boundlist *sublist[MAX_HOLES];
	struct polygondomain *ovortouch();
	struct polygondomain *plgd;
	struct polygondomain *blgd[2];
	struct polygondomain *ptr;
	struct ivertex *wtx;
	int hnum, i, j, aveg;
	GREY g;
	struct ipoint point;

	/*
	 * check that the object to split is sanitary
	 */
	if(woolz_check_obj(comobj, "split_overs.1") != 0)
		return(0);

	/*
	 * check if very small
	 */
	i = area(comobj);
	if (i < 300) {
#ifdef DEBUG
		if (debug_channel != NULL)
			fprintf(debug_channel,"Split_overs: EXIT (too small)\n");
#endif DEBUG
		return(0);
	}

	/*
	 * if there is a hole split it at first
	 */
	obj = comobj;
	mobj = sobj = NULL;
#ifdef DEBUG
	if (debug_channel != NULL)
		fprintf(debug_channel,"Split_overs entry.  Vdom linkcount %d\n",obj->vdom->linkcount);
#endif DEBUG
	bobj = ibound(obj, 1);
	blist = (struct boundlist *) bobj->idom;

	/*
	 * these functions cannot cope with joined disconnected objects
	 * test for these and return/fail with an appropriate message
	 * implement properly when splitlib is revamped
	 */
	if(blist->next != NULL) {
		fprintf(stdout, "Unable to unscramble joined objects.\n");
		freeobj(bobj);
		return(0);
	}
	/*
	 * Object boundary to holey ??
	 */
	getholelist(blist->down, sublist, &hnum);
	if (hnum >= MAX_HOLES) {
		freeobj(bobj);
		fprintf(stderr,"Boundary too noisy for auto-split\007\n");
#ifdef DEBUG
		if (debug_channel != NULL)
			fprintf(debug_channel,"Split_overs: EXIT (too many holes)\n");
#endif DEBUG
		return(0);
	}

	/*
	 * if there is a hole split or mend it at first
	 */
	if(blist->down != NULL){
#ifdef DEBUG
		if (debug_channel != NULL)
			fprintf(debug_channel,"Mendhole called\n");
#endif DEBUG
		aveg = averagegrey(obj, &g);
		if(aveg > 40)
			j = 10;
		else j = 15;
		mobj = mendhole(obj, blist, j);
		if (mobj != NULL) {
#ifdef DEBUG
			if (debug_channel != NULL)
				fprintf(debug_channel,"Mendhole mended holes.  Vdom linkcount %d\n",mobj->vdom->linkcount);
#endif DEBUG
			freeobj(bobj);
			obj = mobj;
			bobj = ibound(obj,1);
			blist = (struct boundlist *) bobj->idom;
		}
	}

	if(blist->down != 0){
		sobj = splithole(obj, blist);
		if (sobj == NULL) {
			freeobj(bobj);
			if (mobj != NULL)
				freeobj(mobj);
#ifdef DEBUG
			if (debug_channel != NULL)
				fprintf(debug_channel,"Split_overs: EXIT (splithole failed)\n");
#endif DEBUG
			return(0);
		} else {
#ifdef DEBUG
			if (debug_channel != NULL)
				fprintf(debug_channel,"Split_overs after splithole.  Vdom linkcount %d\n",obj->vdom->linkcount);
#endif DEBUG
			obj = sobj;
		}   
	}
	freeobj(bobj);
	/*
	 * Check if it is overlaps
	 * Find the four cut points and the curves which can join them
	 * Then split the overlaps along this curves 
	 * Finally restore the grey level
	 */
#ifdef DEBUG
	if (debug_channel != NULL)
		fprintf(debug_channel,"Split_overs before ovortouch\n");
#endif DEBUG
	if((ptr = ovortouch(obj)) == NULL){
		fprintf(stderr,"Decomposition failed\n");
#ifdef DEBUG
		if (debug_channel != NULL)
			fprintf(debug_channel,"Split_overs: EXIT (ovortouch failed)\n");
#endif DEBUG
		if (sobj != NULL)
			freeobj(sobj);
		if (mobj != NULL)
			freeobj(mobj);
		return(0);	
	} else {
		/*
		 * show the procedure of analyse
		 */
		switch(ptr->nvertices){
			default:
				Free(ptr);
				fprintf(stderr,"Decomposition failed\n");
				if (sobj != NULL)
					freeobj(sobj);
				if (mobj != NULL)
					freeobj(mobj);
				return(0);
			case 5:
#ifdef DEBUG
				if (debug_channel != NULL) {
					wtx = ptr->vtx;
					fprintf(debug_channel,"Split_overs: split points from ovortouch -> disoverlap:\n");
					for(i = 0; i < 4; i++,wtx++)
						fprintf(debug_channel,"\t%4d %4d\n",wtx->vtY,wtx->vtX);
				}
#endif DEBUG
				/*
				 * If any pair of cut points is
				 * too close, reject and return
				 */
				if (check_four(ptr)) {
					Free(ptr);
					fprintf(stderr,"Decomposition failed\n");
					if (sobj != NULL)
						freeobj(sobj);
					if (mobj != NULL)
						freeobj(mobj);
					return(0);
				}
#ifdef DEBUGDISPLAY
				/*
				 * show the cut points
				 */
				wtx = ptr->vtx;
				for(i = 0; i < 4; i++){
					point.type = 40;
					point.style = 0;
					point.l = wtx->vtY;
					point.k = wtx->vtX;
					picframe(&point, fr);
					/*circle(10);*/
					wtx++;
				}
				colour(252);
#endif DEBUGDISPLAY
				if(disoverlap(obj, ptr->vtx, new0, new1) == 0) {
					Free(ptr);
					fprintf(stderr,"Decomposition failed\n");
					if (sobj != NULL)
						freeobj(sobj);
					if (mobj != NULL)
						freeobj(mobj);
					return(0);
				}
				break;
		}
		Free(ptr);
		(*new0)->vdom = obj->vdom;
		(*new1)->vdom = obj->vdom;
#ifdef DEBUGDISPLAY1
		subobj[0] = ibound(*new0, 1);
		subobj[1] = ibound(*new1, 1);
		for(i = 0; i < 2; i++){
			sublist[i] = (struct boundlist *) subobj[i]->idom;
			blgd[i] = sublist[i]->poly;
		}
		/*
		 * show the first new object
		 */
		/*
		picframe(*new0, fr);
		colour(252);
		intens(1);
		*/
		colour(col0);
		picpoly(blgd[0], fr);
		intens(1);
		/*
		 * show the second new object
		 */
		/*
		picframe(*new1, fr);
		intens(1);
		colour(252);
		intens(1);
		*/
		colour(col1);
		picpoly(blgd[1], fr);
		intens(1);
		update();
		/*
		 * release space
		 */
		for(i = 0; i < 2; i++){
			freeobj(subobj[i]);
		}
#endif DEBUGDISPLAY1
		if (sobj != NULL)
			freeobj(sobj);
		if (mobj != NULL)
			freeobj(mobj);
#ifdef DEBUG
if (debug_channel != NULL)
fprintf(debug_channel,"Split_overs exit.  Vdom linkcount %d\n",obj->vdom->linkcount);
#endif DEBUG
		/*
		 * if either of the objects produced is bad free both and fail
		 */
		if(woolz_check_obj(*new0, "split_overs.2") ||
				woolz_check_obj(*new1, "split_overs.3")) {
			freeobj(*new0);
			freeobj(*new1);
			return(0);
		}
		return(1);
	}
}


/*
 * check that the set of four split points are
 * really a distinct set.
 */
check_four(ptr)
struct polygondomain *ptr;
{
	register struct ivertex *vtx, *wtx;
	register int i,j;
	vtx = ptr->vtx;
	for (i=0; i<3; i++,vtx++) {
		wtx = vtx+1;
		for (j=i+1; j<4; j++,wtx++)
			if (absdis(vtx,wtx) < 4)
				return(1);
	}
	return(0);
}
