/*
 * fswriteobj.c		Jim Piper		August 1983
 *
 * Routines to convert object structures to sequential data streams.
 * When disc files are intended,
 * any necessary "fseek"-ing must be done elsewhere.
 *
 * The basic philosophy of the sequential form of an object is that
 * sub-objects are ordered in such a way that pointers are unnecessary.
 * See ../doc/filestructs.d for details.
 *
 **********************************************************************
 * VERSION TO EXTRACT GREY-TABLE FROM CRS VME-512 FRAME STORE
 **********************************************************************
 *
 *	Modifications
 *
 *	14 Sep 1989		dcb		MIZAR -> OSK
 *	 4 Dec 1988		CAS		Convert to long woolz names + make compile
 *
 */

#include <stdio.h>
#include <wstruct.h>

int packcheck = 1;

fswriteobj(fp,obj)
FILE *fp;
struct object *obj;
{
	putc((char) obj->type, fp);

	switch (obj->type) {

	case 1:
		writeintervaldomain(fp,obj->idom);
		writevaluetable(fp,obj);
		writepropertylist(fp,obj->plist);
		break;

	case 10:
		writepolygon(fp,obj->idom);
		break;

	case 11:
		writeboundlist(fp,obj->idom);
		break;

	case 13:
		writehisto(fp, obj->idom);
		break;

	case 20:
		writerect(fp,obj->idom);
		break;

	case 30:
	case 31:
		writevector(fp,obj);
		break;

	case 40:
	case 41:
		writepoint(fp,obj);
		break;
	case 70:
		writetext(fp,obj);
		break;

	case 110:
		writepropertylist(fp,obj->plist);
		break;
	default:
		fprintf(stderr,"Can't write object type %d\n",obj->type);
		exit(51);
	}
}


static
writeintervaldomain(fp, itvl)
FILE *fp;
struct intervaldomain *itvl;
{
	register int i, j, nlines;
	register struct intervalline *ivln;

	if (itvl == NULL) {
		putc(0,fp);
		return;
	}
	putc((char) itvl->type, fp);
	fwrite(&(itvl->line1), sizeof(COORD), 4, fp);

	switch (itvl->type) {

	case 1:
		nlines = itvl->lastln - itvl->line1;
		for (i=0; i<=nlines; i++)
			putw(itvl->intvlines[i].nintvs, fp);
		ivln = itvl->intvlines;
		for (i=0; i<=nlines; i++) {
			fwrite(ivln->intvs, sizeof(struct interval), ivln->nintvs, fp);
			ivln++;
		}
		break;

	case 2:
		break;

	default:
		fprintf(stderr,"Can't write domain type %d\n",itvl->type);
		exit(51);
	}
}



static
writepropertylist(fp, plist)
FILE *fp;
struct propertylist *plist;
{
	register char *pl;
	register int i, si;
	if (plist == NULL)
		putc((char) 0, fp);
	else {
		putc((char) 1, fp);
		putw(plist->size, fp);
		pl = (char *) plist;
		pl += sizeof(SMALL);
		si = plist->size - sizeof(SMALL);
		fwrite(pl,1,si,fp);
	}
}



static
writevaluetable(fp, obj)
FILE *fp;
struct object *obj;
{
	struct iwspace iwsp;
	struct gwspace gwsp;
	int packing, type;
	register int i;
	register GREY *g, *gtop, min, max;
	register unsigned char *bg;
	unsigned char gbot[1024];
	short sht;

	if (obj->vdom == NULL) {
		putc(0,fp);
		return;
	}

	type = obj->vdom->type;
	putc((char) (type%10), fp);
	switch (type) {
	case 1:
	case 11:
	case 21:
		if (packcheck != 0) {
			min = 0;
			max = 255;
			initgreyscan(obj,&iwsp,&gwsp);
			while (nextgreyinterval(&iwsp) == 0) {
				g = gwsp.grintptr;
				for (i=0; i<iwsp.colrmn; i++) {
					if (*g < min)
						min = *g;
					else if (*g > max)
						max = *g;
					g++;
				}
			}
			if (min == 0 && max == 255)
				packing = 3;
#ifndef OSK
			else if (min > -32768 && max < 32768)
				packing = 2;
#endif OSK
			else
				packing = 1;
		} else
			packing = 3;
		putc((char) packing, fp);
		fwrite(&(obj->vdom->bckgrnd), sizeof(GREY), 1, fp);
		initgreyscan(obj,&iwsp,&gwsp);
		while(nextgreyinterval(&iwsp) == 0)
			switch (packing) {
			case 1:
				fwrite(gwsp.grintptr, sizeof(GREY), iwsp.colrmn, fp);
				break;
#ifndef OSK
			case 2:
				g = gwsp.grintptr;
				for (i=0; i<iwsp.colrmn; i++) {
					sht = (short) *g;
					fwrite(&sht, sizeof(short), 1, fp);
					g++;
				}
				break;
#endif OSK
			case 3:
				/*
				 * byte packing.  Pack up, use fwrite to
				 * get advantages of buffering.
				 */
				g = gwsp.grintptr;
				gtop = g + iwsp.colrmn;
				bg = gbot;
				while (g < gtop)
					*bg++ = *g++;
				fwrite(gbot, sizeof(char), iwsp.colrmn, fp);
				break;
			}
		break;
	case 2:
	case 12:
	case 22:
		fprintf(stderr,"Can't write float grey tables\n");
		exit(53);
	default:
		fprintf(stderr,"Unknown grey table type %d\n",type);
		exit(53);
	}
}



static
writepolygon(fp, poly)
FILE *fp;
struct polygondomain *poly;
{
	register int nvertices;

	nvertices = poly->nvertices;
	putc((char) poly->type, fp);
	putw(nvertices, fp);
	switch (poly->type) {

	case 1:
		fwrite(poly->vtx, sizeof(struct ivertex), nvertices, fp);
		break;

	case 2:
		fwrite(poly->vtx, sizeof(struct fvertex), nvertices, fp);
		break;
	
	default:
		fprintf(stderr,"Illegal polygon type %d\n",poly->type);
		exit(51);
	}
}


static
writeboundlist (fp,blist)
FILE *fp;
struct boundlist *blist;
{
	if (blist == NULL)
		putc(0,fp);
	else {
		putc(1,fp);
		putc((char)blist->type,fp);
		writeboundlist(fp,blist->next);
		writeboundlist(fp,blist->down);
		putw(blist->wrap,fp);
		writepolygon(fp,blist->poly);
	}
}


static
writerect(fp, rdom)
FILE *fp;
struct irect *rdom;
{
	if (rdom == NULL)
		putc(0,fp);
	else {
		putc(rdom->type,fp);
		switch (rdom->type) {
		case 1:
			fwrite(&(rdom->rangle), sizeof(struct irect) - sizeof(SMALL), 1, fp);
			break;
		case 2:
			fwrite(&(rdom->rangle), sizeof(struct frect) - sizeof(SMALL), 1, fp);
			break;
		}
	}
}
			

static writevector(fp, vec)
struct ivector *vec;
{
	switch (vec->type) {
	case 30:
		fwrite(&vec->style, sizeof(struct ivector) - sizeof(SMALL), 1, fp);
		break;
	case 31:
		fwrite(&vec->style, sizeof(struct fvector) - sizeof(SMALL), 1, fp);
		break;
	}
}
			

static writepoint(fp, pnt)
struct ipoint *pnt;
{
	switch (pnt->type) {
	case 40:
		fwrite(&pnt->style, sizeof(struct ipoint) - sizeof(SMALL), 1, fp);
		break;
	case 41:
		fwrite(&pnt->style, sizeof(struct fpoint) - sizeof(SMALL), 1, fp);
		break;
	}
}


static writehisto(fp, h)
FILE *fp;
struct histogramdomain *h;
{
	putc(h->type, fp);
	fwrite(&h->r, sizeof(struct histogramdomain) - sizeof(SMALL),1,fp);
	fwrite(h->hv, (h->type == 1? sizeof(int): sizeof(float)), h->npoints, fp);
}

static writetext(fp,tobj)
FILE *fp;
struct text *tobj;
{
	struct textdomain *tdom;
	char *tstring,c;
	int i;

	fwrite(tobj->tdom,sizeof(struct textdomain),1,fp);
	tstring = tobj->text;
	tdom = tobj->tdom;
	for(i = 0;i <= tdom->stringlen; i++) {
		c = tstring[i];
		putc(c,fp);
	}
}
