/*
 * nfipobj.c		Jim Piper		14/1/86
 *
 * Construct Woolz object (main type 1, domain type 1, grey-table
 * type 21) from the new-style FIP digitised
 * data passed down the PIA from MIPROC.  This data is produced
 * both by a hardware threshold card and by a software simulation.
 *
 * The format is described by Roy Bailey in the note
 * "FIP DESIGN NOTE (Pixel Mode - Version 2)" dated 12.11.85.
 * Briefly it is as follows:
 *
 * Above threshold pixels are passed as unpacked 16 bit values.
 * RUNS of below threshold pixels ON ONE LINE are passed as a count
 * plus bit 15.  End-of-line is signalled by bit 14; this can occur
 * EITHER in a pixel word OR in a run count.
 *
 * This program currently is making some consistency checks on the
 * data in order to permit debuging of the MIPROC end.
 *
 * Input parameters are (1) address of digitised buffer, (2) length
 * of buffer IN BYTES (as returned e.g. by "readpia()").
 *
 * Modifications
 *
 *	11 Mar 1987		CAS			Turned off debug..
 *   9 Mar 1987		BDP			Fix a negative calloc problem
 *	10 Oct 1986		CAS/GJP		Hit linkcount when setting freeptr
 *	12 Sep 1986		CAS			Includes
 */
#include <stdio.h>
#include <wstruct.h>

#define RUN	0x8000
#define EOLINE 0x4000
#define RUNLENGTH 0x3fff

struct object *nfipobj(fb,fl)
register unsigned short *fb;
{
	register unsigned int u;
	register int kol, nints, lnkols;
	int line, nlines, nkols, runlength;
	int line1, lastln;
	unsigned short *fbwas, *ft;
	struct object *makemain(), *obj;
	struct intervaldomain *idom, *makedomain();
	register struct interval *itvl;
	struct interval *jtvl;
	struct ivtable *v;
	struct valintline *vil;
	register struct valueline *val;
	
	fbwas = fb;
	ft = (unsigned short *) ((char *)fb + fl);
	line = nlines = kol = nkols = lnkols = nints = 0;
	line1 = 0;
	
	/*
	 * find line bounds, number of intervals, check for lost data
	 */
	while (fb < ft) {
		u = *fb++;
		if ((u & RUN) == RUN) {
			runlength = u & RUNLENGTH;
			kol += runlength;
			lnkols += runlength;
		} else {
			/* first pixel in interval */
			kol++;
			lnkols++;
			nints++;
			lastln = line;
			if ((u & EOLINE) == 0) {
				/* skip remaining pixels until next-run or end-of-line */
				while ((*fb & (RUN|EOLINE)) == 0) {
					kol++;
					lnkols++;
					fb++;
				}
				/* If end-of-line but not a below-threshold run then read pixel */
				if ((*fb & RUN) == 0) {
					kol++;
					lnkols++;
					u = *fb++;
				}
			}
		}
		/* handle end-of-line */
		if ((u & EOLINE) == EOLINE) {
#ifdef DEBUG
			if (nkols != lnkols)
				fprintf(stderr,"line %4d kols %4d\n",line,lnkols);
#endif
			nkols = lnkols;
			nlines++;
			line++;
			lnkols = kol = 0;
			if (nints == 0)
				line1++;
		}
	}
#ifdef DEBUG
	fprintf(stderr,"total %d lines\n",nlines);
	fprintf(stderr,"total %d intervals\n",nints);
#endif
	if (nints == 0)
		return(NULL);
	/*
	 * make a suitably sized object
	 */
	idom = makedomain(1,line1,lastln,0,nkols);
	idom->freeptr = (char *) Malloc(nints * sizeof(struct interval));
	obj = makemain(1,idom,NULL,NULL,NULL);
	/*
	 * The following code exracted from "makeivtable" for type 21 grey table.
	 * Allocate space for basic, per line and per interval structures.
	 */
	v = (struct ivtable *) Calloc(sizeof(struct ivtable) +
		(lastln - line1 + 1) * sizeof(struct valintline) +
		nints * sizeof(struct valueline), 1);
	vil = (struct valintline *) (v + 1);
	val = (struct valueline *) (vil + lastln - line1 + 1);
	/*
	 * fill in structure values
	 */
	v->freeptr = (char *) fbwas;
	v->linkcount = 1;		/* So we know its being used */
	v->type = 21;
	v->line1 = line1;
	v->lastln = lastln;
	v->bckgrnd = 0;		/* bckgrnd */
	v->vil = vil;
	v->original = obj;
	vil->vtbint = val;
	obj->vdom = (struct valuetable *) v;

	/*
	 * Rescan data filling intervals structure.
	 */
	itvl = jtvl = (struct interval *) idom->freeptr;
	fb = fbwas;
	line = kol = nints = 0;
	while (fb < ft) {
		u = *fb++;
		if ((u & RUN) == RUN) {
			runlength = u & RUNLENGTH;
			kol += runlength;
		} else {
			/* first pixel in interval */
			vil->nintvs++;
			val->vkol1 = kol;
			val->vlastkl = kol;
			val->values = (GREY *) fb - 1;
			itvl->ileft = kol;
			itvl->iright = kol++;
			nints++;
			if ((u & EOLINE) == 0) {
				/* skip remaining pixels until next-run or end-of-line */
				while ((*fb & (RUN|EOLINE)) == 0) {
					val->vlastkl = kol;
					itvl->iright = kol++;
					fb++;
				}
				/* If end-of-line but not a below-threshold run then read pixel */
				if ((*fb & RUN) == 0) {
					val->vlastkl = kol;
					itvl->iright = kol++;
					u = *fb++;
				}
			}
			/* We must be at the end of the interval here */
			itvl++;
			val++;
		}
		/* handle end-of-line */
		if ((u & EOLINE) == EOLINE) {
			/* ensure that we clear end-of-line bit from pixel data */
			*(fb-1) &= 255;
			if (line >= line1 && line <= lastln) {
				makeinterval(line,idom,nints,jtvl);
				vil++;
				if (line < lastln)
					vil->vtbint = val;
			}
			line++;
			kol = 0;
			nints = 0;
			jtvl = itvl;
		}
	}
	standardidom(idom);
	return(obj);
}
