/************************************************************************/
/*																		*/
/*						polysplit.c										*/
/*																		*/
/************************************************************************/
/*																		*/
/*	Written by:		B. Pinnington										*/
/*																		*/
/*	Purpose:		Test program for woolz function polysplit			*/
/*																		*/
/*	Description:	Allows one or two filenames to be input.		 	*/
/*					There must always be two sources of input to this	*/
/*					module, one of type 1 objects , and the other of	*/
/*					type 10 polygon objects.	If only one filename	*/
/*					is given then the source of other data is assumed 	*/
/*					to be stdin. When both files have been fully read	*/
/*					a split will be performed on a one for one basis	*/
/*					until the smaller file is exhausted. It would 		*/
/*					normally be the case that the files supplied would	*/
/*					be of just one object in length each for polysplit	*/
/*					testing, but the multiple object capability, being	*/
/*					the same as that of extractov allows parallel		*/
/*					automatic validation to be performed.				*/
/*					As the routine is intelligent enough to spot which	*/
/*					file is supplying which sort of input, either		*/
/*					type 1 or type 10 can be piped into the stdin as	*/
/*					required.											*/
/*																		*/
/************************************************************************/
/*																		*/
/*	Modifications:														*/
/*																		*/
/*	05 Nov 1987		BDP			Changed pointer lists to 200 elements	*/
/*																		*/
/************************************************************************/

#include <stdio.h>
#include <wstruct.h>

/*page*/
main(argc,argv)
int argc;
char **argv;
{
	FILE *fp1, *fp2, *fp3;
	struct object *obj, *readobj(), *polysplit(), *objs[200], *obj2s[200], *tobjs[200];
	struct object **pobj, **obj1, **objsptr, **obj2sptr;
	char *str;
	int i,j,fc,count;
	
	/*  N.B.  If unsegmented cells are to be read, then an extra memory board
	will almost certainly be required, even then use of this routine in pipes
	should be controlled */
	
	Alloc_limits(0, 1024 * 512 * sizeof(GREY) );
	setmemalloc(65536);
	fp2 = stdin;
		
	if (argc > 1)
	{
		if ( (fp1 = fopen(argv[1],"r") ) == NULL)
		{
			fprintf( stderr,"POLYSPLIT ERROR: Unable to open file %s\n",argv[1]);
			exit(1);
		}
		else
		{
			if (argc > 2)
			{
				if ((fp2 = fopen(argv[2],"r")) == NULL)
				{
					fprintf(stderr,"POLYSPLIT ERROR: Unable to open file %s\n",argv[2]);
					exit(1);
				}
			}
		}
	}
	else
	{
		fprintf(stderr,"POLYSPLIT ERROR: Too few arguements\n");
		exit(2);
	}
	
	i = j = 0;		/* initialise both array vertices */

	/* !!!! N.B. !!!!! It is important that fp2 is read first when this can
	be stdin, otherwise the stdout file block will get initialised to
	unbuffered. mputc doesn't cope with this and the first character of the 
	writeobj will get lost !! */
	
	while((obj = readobj(fp2)) !=NULL )
		obj2s[j++] = obj;
	if (fp2 == stdin)
		str = "stdin";
	else
		str = argv[2];
	fprintf(stderr,"POLYSPLIT: %d objects read from file %s\n",j,str);

	while((obj = readobj(fp1)) != NULL)
		objs[i++] = obj;
	fprintf(stderr,"POLYSPLIT: %d objects read from file %s\n",i,argv[1]);

	/* now want to work out which array contains type 1 objects. and 
	which contains type 10.  Need to bear in mind possible presence of
	header record, and possible duff array */
	
	objsptr = &objs[0];
	obj2sptr = &obj2s[0];
	while ( i > 0)
	{
		if ( ((*objsptr)->type == 10) || ((*objsptr)->type == 1) )
			break;
		else
		{
			/* reduce the count for this array and increment start pointer */
			i--;
			objsptr++;
		}
	}
	while ( j > 0)
	{
		if ( ((*obj2sptr)->type == 10) || ((*obj2sptr)->type == 1) )
			break;
		else
		{
			j--;
			obj2sptr++;
		}
	}
	fc = i < j ? i : j;			/* how many extracts are we going for */
		
	if ( fc <= 0)
	{
		fprintf(stderr,"POLYSPLIT ERROR: Count zero unable to proceed\n");
		exit(3);
	}
	
	fprintf(stderr,"POLYSPLIT: i = %d, j = %d, fc = %d\n",i,j,fc);
	
	/* at this point we know both arrays contain one or more type 1 or type 10
	objects, now check that they contain different types and identify which
	is which */
	
	if ( ((*objsptr)->type == 1) && ((*obj2sptr)->type == 10) )
	{
		obj1 = objsptr;
		pobj = obj2sptr;
	}
	else
		if ( ((*objsptr)->type == 10) && ((*obj2sptr)->type == 1) )
		{
			pobj = objsptr;
			obj1 = obj2sptr;
		}
		else
		{
			fprintf( stderr,"POLYSPLIT ERROR: Unable to distinguish a type 1 and 10 object source\n");
			exit(4);
		}
					
	while ( fc-- > 0)
	{
		if ((*obj1)->type != 1)
		{
			fprintf(stderr,"POLYSPLIT WARNING: Type %d object ignored\n",(*obj1)->type);
			obj1++;
		}
		else
			if((*pobj)->type != 10)
			{
				fprintf(stderr,"POLYSPLIT WARNING: Type %d polyobj ignored\n",(*pobj)->type);
				pobj++;
			}
			else
			{
				polysplit(*obj1, *pobj, tobjs, &count ); 
				for ( i = 0; i < count; i++)
				{
					if ( tobjs[i] != NULL )
						writeobj(stdout,tobjs[i]);
					freeobj(tobjs[i]);
				}
			}
	}
			
}
