/*
 *	P R O B J T O P O L Y . C --
 *
 *  Written: B. Pinnington.
 *           Image Recognition Systems
 *           720 Birchwood Boulevard
 *           Birchwood Science Park
 *           Warrington
 *           Cheshire
 *           WA3 7PX
 *
 *	Copyright (c) and intellectual property rights Image Recognition Systems
 *
 *  Date:   November 1987.
 *
 *	Purpose: Produce polygon version of input objects
 *
 *	Description: Loops reading objects from stdin and converts them to a
 *				 polygon object by one of 3 methods.
 *			-m	 (default)Produces a medial line polygon.
 *			-b	 Produces an outer level boundary closed polygon.
 *			-s	 Produces a skeleton polygon. Takes the skeleton with a
 *				 high smooth passes value to minimise branching, then takes
 *				 the medial line of the result.
 *
 *	Modifications:
 *
 *	29 Mar 1988		SEAJ	Header + tidy up
 *	23 Nov 1987		BDP		Added help option
 *
 */
#include <stdio.h>
#include <wstruct.h>

struct object *medpoly();

main(argc,argv)
int argc;
char **argv;
{
	struct object *obj, *pobj, *nobj, *readobj(), *ibound(), *unitbound(), *skeleton();
	char option;

	setmemalloc(64 * 1024);
	Alloc_limits( 0, 512 * 1024 * sizeof(GREY) );
	option = 0;
	if (argc > 1)
		if(argv[1][0] == '-')
			option = argv[1][1];

	if ( option == '?' ) {
		fprintf(stderr,"Syntax:     probjtopoly [<opt>]\n");
		fprintf(stderr,"Function:   convert type one object to type 10 polygon\n");
		fprintf(stderr,"Options:\n");
		fprintf(stderr,"    -m      Produce medial axis polygon (default)\n");
		fprintf(stderr,"    -b      Produce boundary polygon \n");
		fprintf(stderr,"    -s      Produce skeletal polygon \n");
		exit(0);
	} else while((obj = readobj(stdin)) != NULL) {
		if (obj->type != 1)
			fprintf(stderr,"OBJTOPOLY: Type %d object, ignored\n",obj->type);
		else {
			switch (option) {
			default:
			case 'm':			/*  median line polygon */
				pobj = medpoly(obj);
				break;
			case 'b':			/*  boundary polygon	*/
				nobj = ibound(obj, 1);
				pobj = unitbound( nobj );	/* creates type 10 object from top level polygon */
				freeobj(nobj);
				break;
			case 's':			/*  skeletal polygon	*/
				nobj = skeleton( obj, 5);	/* 5 is to help smooth off branches */
				pobj = medpoly(nobj);		/* convert type 1 to type 10, rather roughly  */
				freeobj(nobj);
				break;
			}	/* end switch */
			writeobj(stdout,pobj);
			freeobj(pobj);

		}
		freeobj(obj);
	}
}

/*
 *		M E D P O L Y  --  Create medial line polygon from type 1 object
 *
 *	Description:	For any non empty line of a type 1 object will
 *					create a single vertex at the mid point . On a 
 *					multiple interval line, this may not coincide with
 *					an interval. 
 *
 *	Entry:			Interval structured object.
 *
 *	Exit:			Polygon object.
 *
 */
struct object *medpoly(obj)
struct object *obj;
{
	struct object *makemain();
	struct polygondomain *pdom;
	struct intervaldomain *idom;
	register struct interval *iv;
	register struct intervalline *il;
	register struct ivertex *vtex;

	register int i,kol1;
	register int nlines,leftk;

	/* make a polygon whose vertices are just the medial point of each line.
	First thing is to make a polygon with number of lines points.	*/

	idom = obj->idom;
	nlines = idom->lastln - idom->line1 + 1;
	pdom = (struct polygondomain *) Calloc(sizeof(struct polygondomain) +
			sizeof(struct ivertex) * nlines , 1);
	pdom->type = 1;
	pdom->nvertices = nlines;
	pdom->maxvertices = nlines;
	pdom->vtx = (struct ivertex *) (pdom + 1);

	kol1 = idom->kol1;
	il = idom->intvlines;
	vtex = pdom->vtx;
	for ( i = idom->line1; i <= idom->lastln ; i++) {
		if (il->nintvs <= 0)
			pdom->nvertices--;		/* dont produce a vertex for an empty line */
		else {
			iv = il->intvs;
			leftk = iv->ileft;
			iv += il->nintvs - 1;
			vtex->vtX = (iv->iright + leftk)/2 + kol1;
			vtex->vtY = i;
			vtex++;					/* only increment if vertex created */
		}
		il++;
	}
	return( makemain(10,pdom,NULL,NULL,NULL) );
}

