/************************************************************************/
/*																		*/
/*						uncastel.c										*/
/*																		*/
/************************************************************************/
/*																		*/
/*	Written by:		B Pinnington										*/
/*																		*/
/*	Purpose:		Possible correction to FIP digitise castelation		*/
/*					problem												*/
/*																		*/
/*	Description:	Reads objects from stdin , and if type 1 then       */
/*					performs a column correction upon the object by		*/
/*					replacing all even columns with the average of their*/
/*					two neighbours on that line. Resolution is reduced  */
/*					to half therefore, but grey resolution may be		*/
/*					improved by removing castellation effect			*/
/*																		*/
/************************************************************************/
/*																		*/
/*	Modifications:														*/
/*																		*/
/*																		*/
/************************************************************************/

#include <stdio.h>
#include <wstruct.h>

/*page*/
/************************************************************************/
/*																		*/
/*	Function:		uncastel											*/
/*																		*/
/*	Purpose:		Remove castellation by adjacent col averaging		*/
/*																		*/
/*	Description:	Attempts to remove FIP castellation by replacing	*/
/*					grey values in alternate columns with the average of*/
/*					their neighbours on that line						*/
/*																		*/
/*	Entry:			Pointer to type one object, and optional background */
/*					value, if negative then vdom value is used otherwise*/
/*					this value overrides the vdom background			*/
/*																		*/
/*	Exit:			Object received has modified grey table				*/
/*																		*/
/************************************************************************/

uncastel(obj,background)
register struct object *obj;
int background;
{
	register int nkols, startkol, g1, g2;
	register GREY *g;
	static struct iwspace iwsp;
	static struct gwspace gwsp;
	
#ifdef DEBUG
	fprintf(stderr,"UNCASTEL: background = %d\n",background);
#endif DEBUG
	/* first check that object is healthy and contains a vdom */
	
	if( (wzcheckobj(obj) > 0) || ( wzemptyidom(obj->idom) > 0))
		return(1);
	if ( obj->vdom == NULL )
		return(1);
		
	if (background < 0)
		background = obj->vdom->bckgrnd;	/* default off intv edge grey value */	
	initgreyscan(obj,&iwsp, &gwsp );
	while ( nextgreyinterval(&iwsp) == 0 ){
		g = gwsp.grintptr;		/* use register pointer for accessing */
		nkols = iwsp.colrmn;	/* use register int for speed */
		startkol = 1;			/* signal that g[-1] is illegal */
				
		/* skip the first column if its odd so that we always start on an
		even column, otherwise we'd end up with a mess. Do this even if only
		one column, as this will be ignored anyway later  */
		
		if ( (iwsp.colpos % 2) > 0 ){
			g++;
			nkols--;
			startkol = 0;		/* ok now to reference g[-1] */
		}
		
		/* process the whole of the current interval by replacing refeneced
		grey value with average of its neigbours. Must check that first one
		can look at its predecessor, and last one at its successor */
		
		while ( nkols-- > 0){
			if( startkol ){
#ifdef DEBUG
	fprintf(stderr,"UNCASTEL: startkol true, kol = %d\n",iwsp.colpos);
#endif DEBUG
				g1 = background;
			} else
				g1 = g[-1];
			
			/* dec of nkols at loop top sets it to zero if curr grey is last */
			if( nkols == 0 )
				g2 = background;
			else
				g2 = g[1];

			*g = (g1 + g2)/2;
			startkol = 0;		/* set to false every time after first */
			nkols--;			/* dec only once cos of dec at top of loop */
			g += 2;				/* nkols stops this going off the end */
		}
	}			


	return(0);
}

/*page*/
main(argc,argv)
char **argv;
int argc;
{
	struct object *obj, *readobj();
	int background;
	char *str;
		
	background = -1;	
	while(--argc > 0)
	{
		if ((*++argv)[0] == '-') 
		{
			switch( argv[0][1] )
			{
			case '?':
				fprintf(stderr,"Syntax:     uncastel [<opts>]\n");
				fprintf(stderr,"Function:   uncastellate Fip digitised objects\n");
				fprintf(stderr,"Options:\n");
				fprintf(stderr,"    -b=<bckgrnd> Alternative background grey value \n");
				exit(0);
			case 'b':
				if (argv[0][2] == '=')
					background = atoi( &argv[0][3] );
				break;
			default:
				fprintf(stderr,"UNCASTEL: Unrecognised parameter ignored\n");
				break;
			}
		}
		else
			fprintf(stderr,"UNCASTEL: Illegal arguement %s\n",*argv);
	}
	
	while((obj = readobj(stdin)) != NULL) {
		if (obj->type == 1)
		{
			uncastel(obj,background);
			writeobj(stdout,obj);
		}
		else
			writeobj(stdout,obj);
		freeobj(obj);
	}
}
