/*
 *
 *	W Z S U M . C --
 *
 *  Written: B. Pennington
 *           Image Recognition Systems
 *           720 Birchwood Boulevard
 *           Birchwood Science Park
 *           Warrington
 *           Cheshire
 *           WA3 7PX
 *
 *	Copyright (c) and intellectual property rights Image Recognition Systems
 *
 *  Date:    November 1987.
 *
 *	Purpose: Woolz object summary report.	   
 *
 *	Description: Prints an extended summary of most woolz object
 *               types. Output is to stderr.
 *
 *  Modifications:
 * 	20 Nov 1987		BDP		Added output for type 2 rectangles			  
 *							added polylength call						  
 *	18 Nov 1987		BDP		Use Alloc_size rather than direct memctrl.	  
 *							Also modularised and commented better		  
 *	17 Nov 1987		BDP		Changed convex hull area to use hullarea	  
 *							rather than summing ->ba which isnt used!	  
 *																		  
 */

#include <stdio.h>
#include <wstruct.h>

#define DEBUG

extern double polylength();
int elemsize;

/*
 *																		  
 *	Function:		wzsum												  
 *																		  
 *	Purpose:		Write a summary of current object to stderr.		  
 *																		  
 *	Description:	Conditionally produces a Facts like summary of the    
 *					object just tested, but with a little more 			  
 *					information, particularly of memory usage.		 	  
 *																		  
 *	Entry:			Object pointer										  
 *																		  
 *	Exit:			Summary output to stderr							  
 *																		  
 */

wzsum(obj)
struct object *obj;
{

#ifdef DEBUG
	fprintf(stderr,"OBJECT SUMMARY:\n");
#endif DEBUG
	
	if ( wzcheckobj(obj) == 0)
	{
		switch(obj->type)
		{
		case 1:
			sumidom(obj);
			sumvdom(obj);
			
			if (obj->plist != NULL)
				fprintf(stderr,"        Property list length %d\n",obj->plist->size);
			else
				fprintf(stderr,"        Property list NULL\n");
			break;	/* end of type one object */
		case 10:
			sumpoly(obj);
			break;
		case 11:
			break;
		case 12:
			sumcvh(obj);
			break;
		case 13:
			sumhisto(obj);
			break;
		case 20:
			sumrect(obj);
			break;
		case 110:
			if (obj->plist == NULL)
				fprintf(stderr,"Header record NULL\n");
			else
				fprintf(stderr,"Header record length %d\n",obj->plist->size);
			break;
		default:
			fprintf(stderr,"Type %d object\n",obj->type);
			break;
	}	/* end switch */

	fprintf(stderr,"\n");
}
else
	fprintf(stderr,"NULL object\n");
}

/*																		  
 *	Function:		sumidom												  
 *																		  
 *	Purpose:		Write a summary of interval domain to stderr.		  
 *																		  
 *	Description:	prints the following:								  
 *					1. Line and column coordinates of domain extent		  
 *					2. Linearea, Area, Intcount as scanned by woolz		  
 *						functions.										  
 *					3. Size of memory allocated to domain block and 	  
 *						number of intervallines this corresponds to		  
 *					4. Size of memory allocated to interval block and	  
 *						the number of intervals this corresponds to		  
 *																		  
 *	Entry:			Object pointer										  
 *																		  
 *	Exit:			None												  
 *																		  
 */

sumidom(obj)
struct object *obj;
{
	struct intervaldomain *idom;

	idom = obj->idom;
	if (wzemptyidom(idom) > 0)
	{
		fprintf(stderr,"SUMMARY: Empty idom\n");
		return;
	}
	/* for interval structured object print out the domain extents
	then some basic interval derived details using standard functions,
	then look at the two blocks of space allocated printing out their
	size and how many of the relevant structure type that corresponds to */
	
	fprintf(stderr,"Type 1: Idom coordinates  %d,  %d,  %d,   %d\n",
			idom->line1, idom->lastln, idom->kol1, idom->lastkl);
	fprintf(stderr,"        Idom Area %d,  Linearea %d, Intcount  %d\n",
			area(obj),linearea(obj),intcount(idom) );
	elemsize = Alloc_size(idom);
	fprintf(stderr,"        Domain space   %d = linecount of       %d\n",elemsize,
			(elemsize - sizeof(struct intervaldomain)) / sizeof(struct intervalline) );

	elemsize = Alloc_size(idom->freeptr);
	fprintf(stderr,"        Interval space %d = interval count of  %d\n",
			elemsize, elemsize / sizeof(struct interval) );

}

/*
 *																		  
 *	Function:		sumvdom												  
 *																		  
 *	Purpose:		Write a summary of value domain to stderr.			  
 *																		  
 *	Description:	prints the following:								  
 *					1. Value domain type								  
 *					2. Line and column coordinates of domain extent		  
 *					3. Linearea, Area, Intcount as scanned or calculated  
 *						for each type in line							  
 *					4. Size of memory allocated to domain block and 	  
 *						number of lines this corresponds to ( not 21 )	  
 *					5. Size of memory allocated to the grey table and 	  
 *						the number of pixels this corresponds to		  
 *																		  
 *	Entry:			Object pointer										  
 *																		  
 *	Exit:			None												  
 *																		  
 */

sumvdom(obj)
struct object *obj;
{
	struct valuetable *vdom;
	struct rectvaltb *rvdom;
	struct ivtable *ivdom;
	struct valueline *vi;
	struct valintline *il;
	int varea, vlinearea, nints, linecount;
	int i, j;
	int line1, lastln, kol1, lastkl, k1;
	
	vdom = obj->vdom;
	if (vdom == NULL)
		fprintf(stderr,"        Value table NULL\n");
	else
	{
		fprintf(stderr,"        Vdom type %d\n",vdom->type);
		nints = 0;
		vlinearea = 0;
		varea = 0;
		kol1 = 1024;
		lastkl = -1;
		elemsize = Alloc_size(vdom); 	/* size of allocated domain for memory
											calculation */
		switch(vdom->type)
		{
		case 1:
			line1 = vdom->line1;
			lastln = vdom->lastln;
			nints = lastln - line1 + 1;
			vi = *vdom->vtblines;
			
			for (i = 0; i < nints; i++)
			{
				varea += vi->vlastkl - vi->vkol1 + 1;
				if (vi->vkol1 < kol1)
					kol1 = vi->vkol1;
				if (vi->vlastkl > lastkl)
					lastkl = vi->vlastkl;
				vi++;
			}
			vlinearea = varea;
			
			linecount = ( elemsize - sizeof(struct valuetable)) /
						( sizeof(struct valueline) + sizeof(struct valueline *) );
			break;
		case 11:
			rvdom = (struct rectvaltb *)vdom;
			line1 = rvdom->line1;
			lastln = rvdom->lastln;
			kol1 = rvdom->kol1;
			lastkl = rvdom->kol1 + rvdom->width - 1;
			varea = (lastln - line1 + 1) * (rvdom->width);
			vlinearea = varea;
			nints = lastln - line1 + 1;
			linecount = ( elemsize - sizeof(struct valuetable)) / sizeof(GREY *);
			break;
		case 21:
			ivdom = (struct ivtable *)vdom;
			line1 = ivdom->line1;
			lastln = rvdom->lastln;
			il = ivdom->vil;
			for ( i = 0; i < lastln - line1 + 1; i++)
			{
				if (il->nintvs > 0)
				{
					nints += il->nintvs;
					vi = il->vtbint;
					k1 = vi->vkol1;		/* remember start of first interval */
					for ( j = 0; j < il->nintvs; j++)
					{
						varea += vi->vlastkl - vi->vkol1 + 1;
						vi++;
					}
					vi--;		/* go back to last interval for lastkl */
					if(vi->vlastkl > lastkl)
						lastkl = vi->vlastkl;
					if(k1 < kol1)
						kol1 = k1;
					vlinearea += vi->vlastkl - k1 + 1;
					il++;
				}
			}											

			break;
		default:
			fprintf(stderr,"        Vdom type %d unknown\n");
		}	/* end switch */
		
		fprintf(stderr,"        Vdom coordinates  %d,  %d,  %d,  %d\n",
				line1, lastln, kol1, lastkl);
		fprintf(stderr,"        Vdom Area %d,  Linearea %d, Intcount %d\n",
				varea, vlinearea, nints);

		/* now print size of value domain memory and the number of lines
		this corresponds to, cant calcualate this for type 21 though */
		
		fprintf(stderr,"        Vdom size %d ", elemsize);
		if (vdom->type != 21)
			fprintf(stderr,"= %d lines\n", linecount);
		else
			fprintf(stderr,"\n");
			
		/* now print out grey table size in bytes and as usable element count */
		
		elemsize = Alloc_size(vdom->freeptr);
		fprintf(stderr,"        Grey table size of %d = %d GREY elements\n",
			elemsize, elemsize/sizeof(GREY) );
	}
}

/*
 *	Function:		sumpdom												  
 *																		  
 *	Purpose:		Write a summary of a polygon to stderr.				  
 *																		  
 *	Description:	prints the following:								  
 *					1. Type , vertex count, maximum vertex count		  
 *					2. Polygon length calculated by polylength			  
 *					3. Size of memory allocated to domain block and 	  
 *						number of vertices this corresponds to			  
 *																		  
 *	Entry:			Object pointer										  
 *																		  
 *	Exit:			None												  
 *																		  
 */

sumpoly(obj)
struct object *obj;
{
	struct polygondomain *pdom;
	int vertexsize;
	
	if (wzemptypdom(obj->idom) > 0)
		fprintf(stderr,"SUMMARY: Empty pdom\n");
	else
	{				
		pdom = (struct polygondomain *) obj->idom;
		elemsize = Alloc_size(pdom);
		if (pdom->type == 1)
			vertexsize = sizeof(struct ivertex);
		else
			vertexsize = sizeof(struct fvertex);
		fprintf(stderr,"Polygon: Type %d  count is %d, maximum is %d, length is %8.4f\n",
				pdom->type, pdom->nvertices, pdom->maxvertices, polylength(pdom) );
		fprintf(stderr,"         Domain space %d = %d vertices \n",
				elemsize, (elemsize - sizeof(struct polygondomain))/
				vertexsize);
	}
}

/*																		  
 *	Function:		sumcvh												  
 *																		  
 *	Purpose:		Write a summary of a convex hull domain to stderr.	  
 *																		  
 *	Description:	prints the following:								  
 *					1. Type , chord count, significant chord count		  
 *					2. Hull perimeter length, hull area as calculated	  
 *						by woolz functions.								  
 *					3. Size of memory allocated to domain block and 	  
 *						number of chords this corresponds to			  
 *																		  
 *	Entry:			Object pointer										  
 *																		  
 *	Exit:			None												  
 *																		  
 */

sumcvh(obj)
struct object *obj;
{
	struct cvhdom *cdom;
	struct chord *chrd;
	int barea, clen;

	sumpoly(obj);

	cdom = (struct cvhdom *) obj->vdom;
	if ( cdom == NULL )
		fprintf(stderr,"SUMMARY: NULL cvhdom\n");
	else
	{
		fprintf(stderr,"Convex Hull: Type %d,  Chord count %d, Sig. count %d\n",
				cdom->type, cdom->nchords, cdom->nsigchords);
		barea = hullarea(obj, &clen);
		fprintf(stderr,"             Total chord lengths %d, hull area %d\n",  clen, barea);
		elemsize = Alloc_size(cdom);
		fprintf(stderr,"             Domain space %d = %d chords \n",
				elemsize, (elemsize - sizeof(struct cvhdom))/ sizeof(struct chord));
	}
}
 
/*
 *	Function:		sumhisto											  
 *																		  
 *	Purpose:		Write a summary of a histogram domain to stderr.	  
 *																		  
 *	Description:	prints the following:								  
 *					1. Row and column origin, count of points			  
 *					2. Size of memory allocated to domain block and 	  
 *						number of histogram points this corresponds to	  
 *																		  
 *	Entry:			Object pointer										  
 *																		  
 *	Exit:			None												  
 *																		  
 */

sumhisto(obj)
struct object *obj;
{
	struct histogramdomain *hdom;

	hdom = (struct histogramdomain *)obj->idom;
	if ( hdom == NULL )
		fprintf(stderr,"SUMMARY: NULL hdom\n");
	else
		fprintf(stderr,"Histogram: column %d, row %d, count %d \n",
				hdom->k, hdom->l, hdom->npoints);
		elemsize = Alloc_size(hdom);
		fprintf(stderr,"           Domain space %d = %d points \n",
				elemsize, (elemsize - sizeof(struct histogramdomain))/ sizeof(int));
}

/*																		  
 *	Function:		sumrect												  
 *																		  
 *	Purpose:		Write a summary of a rectangle domain to stderr.	  
 *																		  
 *	Description:	prints the following:								  
 *					1. Angle of orientation in radians					  
 *					2. Coordinates of corners of the rectangle.			  
 *																		  
 *	Entry:			Object pointer										  
 *																		  
 *	Exit:			None												  
 *																		  
 */

sumrect(obj)
struct object *obj;
{
	struct irect *rdom;
	struct frect *frdom;

	rdom = (struct irect *)obj->idom;
	if ( rdom == NULL )
		fprintf(stderr,"SUMMARY: NULL rdom\n");
	else
	{
		switch(rdom->type ){
		case 1:
			fprintf(stderr,"Rectangle: Type 1, angle in radians is %8.4f\n",rdom->rangle);
			fprintf(stderr,"           corners at  %d,%d : %d,%d : %d,%d : %d,%d\n",
					rdom->irl[0],rdom->irk[0], rdom->irl[1], rdom->irk[1],
					rdom->irl[2],rdom->irk[2], rdom->irl[3], rdom->irk[3]);
			break;
		case 2:
			frdom = (struct frect *)rdom;
			fprintf(stderr,"Rectangle: Type 2, angle in radians is %8.4f\n",frdom->rangle);
			fprintf(stderr,"           corners at  %8.4f,%8.4f : %8.4f,%8.4f \n",
					frdom->frl[0],frdom->frk[0], frdom->frl[1], frdom->frk[1]);
			fprintf(stderr,"                       %8.4f,%8.4f : %8.4f,%8.4f \n",
					frdom->frl[2],frdom->frk[2], frdom->frl[3], frdom->frk[3]);
			break;
		}
	}
}

